# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from google.protobuf.json_format import MessageToJson
from google.protobuf.message import Message

from feast.base_feature_view import BaseFeatureView
from feast.data_source import DataSource
from feast.entity import Entity
from feast.feature_service import FeatureService
from feast.feature_view import FeatureView
from feast.infra.infra_object import Infra
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.permissions.permission import Permission
from feast.project import Project
from feast.project_metadata import ProjectMetadata
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.Entity_pb2 import Entity as EntityProto
from feast.protos.feast.core.FeatureService_pb2 import (
    FeatureService as FeatureServiceProto,
)
from feast.protos.feast.core.FeatureView_pb2 import FeatureView as FeatureViewProto
from feast.protos.feast.core.OnDemandFeatureView_pb2 import (
    OnDemandFeatureView as OnDemandFeatureViewProto,
)
from feast.protos.feast.core.Permission_pb2 import Permission as PermissionProto
from feast.protos.feast.core.Project_pb2 import Project as ProjectProto
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.protos.feast.core.SavedDataset_pb2 import SavedDataset as SavedDatasetProto
from feast.protos.feast.core.StreamFeatureView_pb2 import (
    StreamFeatureView as StreamFeatureViewProto,
)
from feast.saved_dataset import SavedDataset, ValidationReference
from feast.stream_feature_view import StreamFeatureView
from feast.transformation.pandas_transformation import PandasTransformation
from feast.transformation.substrait_transformation import SubstraitTransformation


class BaseRegistry(ABC):
    """
    The interface that Feast uses to apply, list, retrieve, and delete Feast objects (e.g. entities,
    feature views, and data sources).
    """

    # Entity operations
    @abstractmethod
    def apply_entity(self, entity: Entity, project: str, commit: bool = True):
        """
        Registers a single entity with Feast

        Args:
            entity: Entity that will be registered
            project: Feast project that this entity belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def delete_entity(self, name: str, project: str, commit: bool = True):
        """
        Deletes an entity or raises an exception if not found.

        Args:
            name: Name of entity
            project: Feast project that this entity belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_entity(self, name: str, project: str, allow_cache: bool = False) -> Entity:
        """
        Retrieves an entity.

        Args:
            name: Name of entity
            project: Feast project that this entity belongs to
            allow_cache: Whether to allow returning this entity from a cached registry

        Returns:
            Returns either the specified entity, or raises an exception if
            none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_entities(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[Entity]:
        """
        Retrieve a list of entities from the registry

        Args:
            allow_cache: Whether to allow returning entities from a cached registry
            project: Filter entities based on project name
            tags: Filter by tags

        Returns:
            List of entities
        """
        raise NotImplementedError

    # Data source operations
    @abstractmethod
    def apply_data_source(
        self, data_source: DataSource, project: str, commit: bool = True
    ):
        """
        Registers a single data source with Feast

        Args:
            data_source: A data source that will be registered
            project: Feast project that this data source belongs to
            commit: Whether to immediately commit to the registry
        """
        raise NotImplementedError

    @abstractmethod
    def delete_data_source(self, name: str, project: str, commit: bool = True):
        """
        Deletes a data source or raises an exception if not found.

        Args:
            name: Name of data source
            project: Feast project that this data source belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_data_source(
        self, name: str, project: str, allow_cache: bool = False
    ) -> DataSource:
        """
        Retrieves a data source.

        Args:
            name: Name of data source
            project: Feast project that this data source belongs to
            allow_cache: Whether to allow returning this data source from a cached registry

        Returns:
            Returns either the specified data source, or raises an exception if none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_data_sources(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[DataSource]:
        """
        Retrieve a list of data sources from the registry

        Args:
            project: Filter data source based on project name
            allow_cache: Whether to allow returning data sources from a cached registry
            tags: Filter by tags

        Returns:
            List of data sources
        """
        raise NotImplementedError

    # Feature service operations
    @abstractmethod
    def apply_feature_service(
        self, feature_service: FeatureService, project: str, commit: bool = True
    ):
        """
        Registers a single feature service with Feast

        Args:
            feature_service: A feature service that will be registered
            project: Feast project that this entity belongs to
        """
        raise NotImplementedError

    @abstractmethod
    def delete_feature_service(self, name: str, project: str, commit: bool = True):
        """
        Deletes a feature service or raises an exception if not found.

        Args:
            name: Name of feature service
            project: Feast project that this feature service belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_feature_service(
        self, name: str, project: str, allow_cache: bool = False
    ) -> FeatureService:
        """
        Retrieves a feature service.

        Args:
            name: Name of feature service
            project: Feast project that this feature service belongs to
            allow_cache: Whether to allow returning this feature service from a cached registry

        Returns:
            Returns either the specified feature service, or raises an exception if
            none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_feature_services(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[FeatureService]:
        """
        Retrieve a list of feature services from the registry

        Args:
            allow_cache: Whether to allow returning entities from a cached registry
            project: Filter entities based on project name
            tags: Filter by tags

        Returns:
            List of feature services
        """
        raise NotImplementedError

    # Feature view operations
    @abstractmethod
    def apply_feature_view(
        self, feature_view: BaseFeatureView, project: str, commit: bool = True
    ):
        """
        Registers a single feature view with Feast

        Args:
            feature_view: Feature view that will be registered
            project: Feast project that this feature view belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def delete_feature_view(self, name: str, project: str, commit: bool = True):
        """
        Deletes a feature view or raises an exception if not found.

        Args:
            name: Name of feature view
            project: Feast project that this feature view belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    # stream feature view operations
    @abstractmethod
    def get_stream_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> StreamFeatureView:
        """
        Retrieves a stream feature view.

        Args:
            name: Name of stream feature view
            project: Feast project that this feature view belongs to
            allow_cache: Allow returning feature view from the cached registry

        Returns:
            Returns either the specified feature view, or raises an exception if
            none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_stream_feature_views(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[StreamFeatureView]:
        """
        Retrieve a list of stream feature views from the registry

        Args:
            project: Filter stream feature views based on project name
            allow_cache: Whether to allow returning stream feature views from a cached registry
            tags: Filter by tags

        Returns:
            List of stream feature views
        """
        raise NotImplementedError

    # on demand feature view operations
    @abstractmethod
    def get_on_demand_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> OnDemandFeatureView:
        """
        Retrieves an on demand feature view.

        Args:
            name: Name of on demand feature view
            project: Feast project that this on demand feature view belongs to
            allow_cache: Whether to allow returning this on demand feature view from a cached registry

        Returns:
            Returns either the specified on demand feature view, or raises an exception if
            none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_on_demand_feature_views(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[OnDemandFeatureView]:
        """
        Retrieve a list of on demand feature views from the registry

        Args:
            project: Filter on demand feature views based on project name
            allow_cache: Whether to allow returning on demand feature views from a cached registry
            tags: Filter by tags

        Returns:
            List of on demand feature views
        """
        raise NotImplementedError

    # regular feature view operations
    @abstractmethod
    def get_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> FeatureView:
        """
        Retrieves a feature view.

        Args:
            name: Name of feature view
            project: Feast project that this feature view belongs to
            allow_cache: Allow returning feature view from the cached registry

        Returns:
            Returns either the specified feature view, or raises an exception if
            none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_feature_views(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[FeatureView]:
        """
        Retrieve a list of feature views from the registry

        Args:
            allow_cache: Allow returning feature views from the cached registry
            project: Filter feature views based on project name
            tags: Filter by tags

        Returns:
            List of feature views
        """
        raise NotImplementedError

    @abstractmethod
    def get_any_feature_view(
        self, name: str, project: str, allow_cache: bool = False
    ) -> BaseFeatureView:
        """
        Retrieves a feature view of any type.

        Args:
            name: Name of feature view
            project: Feast project that this feature view belongs to
            allow_cache: Allow returning feature view from the cached registry

        Returns:
            Returns either the specified feature view, or raises an exception if
            none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_all_feature_views(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[BaseFeatureView]:
        """
        Retrieve a list of feature views of all types from the registry

        Args:
            allow_cache: Allow returning feature views from the cached registry
            project: Filter feature views based on project name
            tags: Filter by tags

        Returns:
            List of feature views
        """
        raise NotImplementedError

    @abstractmethod
    def apply_materialization(
        self,
        feature_view: Union[FeatureView, OnDemandFeatureView],
        project: str,
        start_date: datetime,
        end_date: datetime,
        commit: bool = True,
    ):
        """
        Updates materialization intervals tracked for a single feature view in Feast

        Args:
            feature_view: Feature view that will be updated with an additional materialization interval tracked
            project: Feast project that this feature view belongs to
            start_date (datetime): Start date of the materialization interval to track
            end_date (datetime): End date of the materialization interval to track
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    # Saved dataset operations
    @abstractmethod
    def apply_saved_dataset(
        self,
        saved_dataset: SavedDataset,
        project: str,
        commit: bool = True,
    ):
        """
        Stores a saved dataset metadata with Feast

        Args:
            saved_dataset: SavedDataset that will be added / updated to registry
            project: Feast project that this dataset belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_saved_dataset(
        self, name: str, project: str, allow_cache: bool = False
    ) -> SavedDataset:
        """
        Retrieves a saved dataset.

        Args:
            name: Name of dataset
            project: Feast project that this dataset belongs to
            allow_cache: Whether to allow returning this dataset from a cached registry

        Returns:
            Returns either the specified SavedDataset, or raises an exception if
            none is found
        """
        raise NotImplementedError

    def delete_saved_dataset(self, name: str, project: str, commit: bool = True):
        """
        Delete a saved dataset.

        Args:
            name: Name of dataset
            project: Feast project that this dataset belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def list_saved_datasets(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[SavedDataset]:
        """
        Retrieves a list of all saved datasets in specified project

        Args:
            project: Feast project
            allow_cache: Whether to allow returning this dataset from a cached registry
            tags: Filter by tags

        Returns:
            Returns the list of SavedDatasets
        """
        raise NotImplementedError

    # Validation reference operations
    @abstractmethod
    def apply_validation_reference(
        self,
        validation_reference: ValidationReference,
        project: str,
        commit: bool = True,
    ):
        """
        Persist a validation reference

        Args:
            validation_reference: ValidationReference that will be added / updated to registry
            project: Feast project that this dataset belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def delete_validation_reference(self, name: str, project: str, commit: bool = True):
        """
        Deletes a validation reference or raises an exception if not found.

        Args:
            name: Name of validation reference
            project: Feast project that this object belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_validation_reference(
        self, name: str, project: str, allow_cache: bool = False
    ) -> ValidationReference:
        """
        Retrieves a validation reference.

        Args:
            name: Name of dataset
            project: Feast project that this dataset belongs to
            allow_cache: Whether to allow returning this dataset from a cached registry

        Returns:
            Returns either the specified ValidationReference, or raises an exception if
            none is found
        """
        raise NotImplementedError

    # TODO: Needs to be implemented.
    def list_validation_references(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[ValidationReference]:
        """
        Retrieve a list of validation references from the registry

        Args:
            project: Filter validation references based on project name
            allow_cache: Allow returning validation references from the cached registry
            tags: Filter by tags

        Returns:
            List of request validation references
        """
        raise NotImplementedError

    @abstractmethod
    def list_project_metadata(
        self, project: str, allow_cache: bool = False
    ) -> List[ProjectMetadata]:
        """
        Retrieves project metadata

        Args:
            project: Filter metadata based on project name
            allow_cache: Allow returning feature views from the cached registry

        Returns:
            List of project metadata
        """
        raise NotImplementedError

    @abstractmethod
    def get_project_metadata(self, project: str, key: str) -> Optional[str]:
        """
        Retrieves a custom project metadata value by key.

        Args:
            project: Feast project name
            key: Metadata key

        Returns:
            The metadata value as a string, or None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def update_infra(self, infra: Infra, project: str, commit: bool = True):
        """
        Updates the stored Infra object.

        Args:
            infra: The new Infra object to be stored.
            project: Feast project that the Infra object refers to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_infra(self, project: str, allow_cache: bool = False) -> Infra:
        """
        Retrieves the stored Infra object.

        Args:
            project: Feast project that the Infra object refers to
            allow_cache: Whether to allow returning this entity from a cached registry

        Returns:
            The stored Infra object.
        """
        raise NotImplementedError

    @abstractmethod
    def apply_user_metadata(
        self,
        project: str,
        feature_view: BaseFeatureView,
        metadata_bytes: Optional[bytes],
    ): ...

    @abstractmethod
    def get_user_metadata(
        self, project: str, feature_view: BaseFeatureView
    ) -> Optional[bytes]: ...

    # Permission operations
    @abstractmethod
    def apply_permission(
        self, permission: Permission, project: str, commit: bool = True
    ):
        """
        Registers a single permission with Feast

        Args:
            permission: A permission that will be registered
            project: Feast project that this permission belongs to
            commit: Whether to immediately commit to the registry
        """
        raise NotImplementedError

    @abstractmethod
    def delete_permission(self, name: str, project: str, commit: bool = True):
        """
        Deletes a permission or raises an exception if not found.

        Args:
            name: Name of permission
            project: Feast project that this permission belongs to
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_permission(
        self, name: str, project: str, allow_cache: bool = False
    ) -> Permission:
        """
        Retrieves a permission.

        Args:
            name: Name of permission
            project: Feast project that this permission belongs to
            allow_cache: Whether to allow returning this permission from a cached registry

        Returns:
            Returns either the specified permission, or raises an exception if none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_permissions(
        self,
        project: str,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[Permission]:
        """
        Retrieve a list of permissions from the registry

        Args:
            project: Filter permission based on project name
            allow_cache: Whether to allow returning permissions from a cached registry

        Returns:
            List of permissions
        """
        raise NotImplementedError

    @abstractmethod
    def apply_project(
        self,
        project: Project,
        commit: bool = True,
    ):
        """
        Registers a project with Feast

        Args:
            project: A project that will be registered
            commit: Whether to immediately commit to the registry
        """
        raise NotImplementedError

    @abstractmethod
    def delete_project(
        self,
        name: str,
        commit: bool = True,
    ):
        """
        Deletes a project or raises an ProjectNotFoundException exception if not found.

        Args:
            project: Feast project name that needs to be deleted
            commit: Whether the change should be persisted immediately
        """
        raise NotImplementedError

    @abstractmethod
    def get_project(
        self,
        name: str,
        allow_cache: bool = False,
    ) -> Project:
        """
        Retrieves a project.

        Args:
            name: Feast project name
            allow_cache: Whether to allow returning this permission from a cached registry

        Returns:
            Returns either the specified project, or raises ProjectObjectNotFoundException exception if none is found
        """
        raise NotImplementedError

    @abstractmethod
    def list_projects(
        self,
        allow_cache: bool = False,
        tags: Optional[dict[str, str]] = None,
    ) -> List[Project]:
        """
        Retrieve a list of projects from the registry

        Args:
            allow_cache: Whether to allow returning permissions from a cached registry

        Returns:
            List of project
        """
        raise NotImplementedError

    @abstractmethod
    def proto(self) -> RegistryProto:
        """
        Retrieves a proto version of the registry.

        Returns:
            The registry proto object.
        """
        raise NotImplementedError

    @abstractmethod
    def commit(self):
        """Commits the state of the registry cache to the remote registry store."""
        raise NotImplementedError

    @abstractmethod
    def refresh(self, project: Optional[str] = None):
        """Refreshes the state of the registry cache by fetching the registry state from the remote registry store."""
        raise NotImplementedError

    # Lineage operations
    def get_registry_lineage(
        self,
        project: str,
        allow_cache: bool = False,
        filter_object_type: Optional[str] = None,
        filter_object_name: Optional[str] = None,
    ) -> tuple[List[Any], List[Any]]:
        """
        Get complete registry lineage with relationships and indirect relationships.
        Args:
            project: Feast project name
            allow_cache: Whether to allow returning data from a cached registry
            filter_object_type: Optional filter by object type (dataSource, entity, featureView, featureService)
            filter_object_name: Optional filter by object name
        Returns:
            Tuple of (direct_relationships, indirect_relationships)
        """
        from feast.lineage.registry_lineage import RegistryLineageGenerator

        # Create a registry proto with all objects
        registry_proto = self._build_registry_proto(project, allow_cache)

        # Generate lineage
        lineage_generator = RegistryLineageGenerator()
        relationships, indirect_relationships = lineage_generator.generate_lineage(
            registry_proto
        )

        # Apply filtering if specified
        if filter_object_type and filter_object_name:
            relationships = [
                rel
                for rel in relationships
                if (
                    (
                        rel.source.type.value == filter_object_type
                        and rel.source.name == filter_object_name
                    )
                    or (
                        rel.target.type.value == filter_object_type
                        and rel.target.name == filter_object_name
                    )
                )
            ]
            indirect_relationships = [
                rel
                for rel in indirect_relationships
                if (
                    (
                        rel.source.type.value == filter_object_type
                        and rel.source.name == filter_object_name
                    )
                    or (
                        rel.target.type.value == filter_object_type
                        and rel.target.name == filter_object_name
                    )
                )
            ]

        return relationships, indirect_relationships

    def get_object_relationships(
        self,
        project: str,
        object_type: str,
        object_name: str,
        include_indirect: bool = False,
        allow_cache: bool = False,
    ) -> List[Any]:
        """
        Get relationships for a specific object.
        Args:
            project: Feast project name
            object_type: Type of object (dataSource, entity, featureView, featureService, feature)
            object_name: Name of the object
            include_indirect: Whether to include indirect relationships
            allow_cache: Whether to allow returning data from a cached registry
        Returns:
            List of relationships involving the specified object
        """
        from feast.lineage.registry_lineage import (
            RegistryLineageGenerator,
        )

        registry_proto = self._build_registry_proto(project, allow_cache)
        lineage_generator = RegistryLineageGenerator()
        relationships = lineage_generator.get_object_relationships(
            registry_proto, object_type, object_name, include_indirect=include_indirect
        )
        return relationships

    def _build_registry_proto(
        self, project: str, allow_cache: bool = False
    ) -> RegistryProto:
        """Helper method to build a registry proto with all objects."""
        registry = RegistryProto()

        # Add all entities
        entities = self.list_entities(project=project, allow_cache=allow_cache)
        for entity in entities:
            registry.entities.append(entity.to_proto())

        # Add all data sources
        data_sources = self.list_data_sources(project=project, allow_cache=allow_cache)
        for data_source in data_sources:
            registry.data_sources.append(data_source.to_proto())

        # Add all feature views
        feature_views = self.list_feature_views(
            project=project, allow_cache=allow_cache
        )
        for feature_view in feature_views:
            registry.feature_views.append(feature_view.to_proto())

        # Add all stream feature views
        stream_feature_views = self.list_stream_feature_views(
            project=project, allow_cache=allow_cache
        )
        for stream_feature_view in stream_feature_views:
            registry.stream_feature_views.append(stream_feature_view.to_proto())

        # Add all on-demand feature views
        on_demand_feature_views = self.list_on_demand_feature_views(
            project=project, allow_cache=allow_cache
        )
        for on_demand_feature_view in on_demand_feature_views:
            registry.on_demand_feature_views.append(on_demand_feature_view.to_proto())

        # Add all feature services
        feature_services = self.list_feature_services(
            project=project, allow_cache=allow_cache
        )
        for feature_service in feature_services:
            registry.feature_services.append(feature_service.to_proto())

        return registry

    @staticmethod
    def _message_to_sorted_dict(message: Message) -> Dict[str, Any]:
        return json.loads(MessageToJson(message, sort_keys=True))

    def to_dict(self, project: str) -> Dict[str, List[Any]]:
        """Returns a dictionary representation of the registry contents for the specified project.
        For each list in the dictionary, the elements are sorted by name, so this
        method can be used to compare two registries.
        Args:
            project: Feast project to convert to a dict
        """
        registry_dict: Dict[str, Any] = defaultdict(list)
        registry_dict["project"] = project
        for project_metadata in sorted(self.list_project_metadata(project=project)):
            registry_dict["projectMetadata"].append(
                self._message_to_sorted_dict(project_metadata.to_proto())
            )
        for data_source in sorted(
            self.list_data_sources(project=project), key=lambda ds: ds.name
        ):
            registry_dict["dataSources"].append(
                self._message_to_sorted_dict(data_source.to_proto())
            )
        for entity in sorted(
            self.list_entities(project=project),
            key=lambda entity: entity.name,
        ):
            registry_dict["entities"].append(
                self._message_to_sorted_dict(entity.to_proto())
            )
        for feature_view in sorted(
            self.list_feature_views(project=project),
            key=lambda feature_view: feature_view.name,
        ):
            registry_dict["featureViews"].append(
                self._message_to_sorted_dict(feature_view.to_proto())
            )
        for feature_service in sorted(
            self.list_feature_services(project=project),
            key=lambda feature_service: feature_service.name,
        ):
            registry_dict["featureServices"].append(
                self._message_to_sorted_dict(feature_service.to_proto())
            )
        for on_demand_feature_view in sorted(
            self.list_on_demand_feature_views(project=project),
            key=lambda on_demand_feature_view: on_demand_feature_view.name,
        ):
            odfv_dict = self._message_to_sorted_dict(on_demand_feature_view.to_proto())
            # We are logging a warning because the registry object may be read from a proto that is not updated
            # i.e., we have to submit dual writes but in order to ensure the read behavior succeeds we have to load
            # both objects to compare any changes in the registry
            warnings.warn(
                "We will be deprecating the usage of spec.userDefinedFunction in a future release please upgrade cautiously.",
                DeprecationWarning,
            )
            if on_demand_feature_view.feature_transformation:
                if isinstance(
                    on_demand_feature_view.feature_transformation, PandasTransformation
                ):
                    if "userDefinedFunction" not in odfv_dict["spec"]:
                        odfv_dict["spec"]["userDefinedFunction"] = {}
                    odfv_dict["spec"]["userDefinedFunction"]["body"] = (
                        on_demand_feature_view.feature_transformation.udf_string
                    )
                    odfv_dict["spec"]["featureTransformation"]["userDefinedFunction"][
                        "body"
                    ] = on_demand_feature_view.feature_transformation.udf_string
                elif isinstance(
                    on_demand_feature_view.feature_transformation,
                    SubstraitTransformation,
                ):
                    odfv_dict["spec"]["featureTransformation"]["substraitPlan"][
                        "body"
                    ] = on_demand_feature_view.feature_transformation.substrait_plan
                else:
                    odfv_dict["spec"]["featureTransformation"]["userDefinedFunction"][
                        "body"
                    ] = None
                    odfv_dict["spec"]["featureTransformation"]["substraitPlan"][
                        "body"
                    ] = None
                registry_dict["onDemandFeatureViews"].append(odfv_dict)
        for stream_feature_view in sorted(
            self.list_stream_feature_views(project=project),
            key=lambda stream_feature_view: stream_feature_view.name,
        ):
            sfv_dict = self._message_to_sorted_dict(stream_feature_view.to_proto())

            sfv_dict["spec"]["userDefinedFunction"]["body"] = (
                stream_feature_view.udf_string
            )
            registry_dict["streamFeatureViews"].append(sfv_dict)

        for saved_dataset in sorted(
            self.list_saved_datasets(project=project), key=lambda item: item.name
        ):
            registry_dict["savedDatasets"].append(
                self._message_to_sorted_dict(saved_dataset.to_proto())
            )
        for infra_object in sorted(self.get_infra(project=project).infra_objects):
            registry_dict["infra"].append(
                self._message_to_sorted_dict(infra_object.to_proto())
            )
        for permission in sorted(
            self.list_permissions(project=project), key=lambda ds: ds.name
        ):
            registry_dict["permissions"].append(
                self._message_to_sorted_dict(permission.to_proto())
            )

        return registry_dict

    @staticmethod
    def deserialize_registry_values(serialized_proto, feast_obj_type) -> Any:
        if feast_obj_type == Entity:
            return EntityProto.FromString(serialized_proto)
        if feast_obj_type == SavedDataset:
            return SavedDatasetProto.FromString(serialized_proto)
        if feast_obj_type == FeatureView:
            return FeatureViewProto.FromString(serialized_proto)
        if feast_obj_type == StreamFeatureView:
            return StreamFeatureViewProto.FromString(serialized_proto)
        if feast_obj_type == OnDemandFeatureView:
            return OnDemandFeatureViewProto.FromString(serialized_proto)
        if feast_obj_type == FeatureService:
            return FeatureServiceProto.FromString(serialized_proto)
        if feast_obj_type == Permission:
            return PermissionProto.FromString(serialized_proto)
        if feast_obj_type == Project:
            return ProjectProto.FromString(serialized_proto)
        if issubclass(feast_obj_type, DataSource):
            return DataSourceProto.FromString(serialized_proto)
        return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/Transformation.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1f\x66\x65\x61st/core/Transformation.proto\x12\nfeast.core\"F\n\x15UserDefinedFunctionV2\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x62ody\x18\x02 \x01(\x0c\x12\x11\n\tbody_text\x18\x03 \x01(\t\"\xba\x01\n\x17\x46\x65\x61tureTransformationV2\x12\x42\n\x15user_defined_function\x18\x01 \x01(\x0b\x32!.feast.core.UserDefinedFunctionV2H\x00\x12I\n\x18substrait_transformation\x18\x02 \x01(\x0b\x32%.feast.core.SubstraitTransformationV2H\x00\x42\x10\n\x0etransformation\"J\n\x19SubstraitTransformationV2\x12\x16\n\x0esubstrait_plan\x18\x01 \x01(\x0c\x12\x15\n\ribis_function\x18\x02 \x01(\x0c\x42_\n\x10\x66\x65\x61st.proto.coreB\x1a\x46\x65\x61tureTransformationProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.Transformation_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\032FeatureTransformationProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_USERDEFINEDFUNCTIONV2']._serialized_start=47
  _globals['_USERDEFINEDFUNCTIONV2']._serialized_end=117
  _globals['_FEATURETRANSFORMATIONV2']._serialized_start=120
  _globals['_FEATURETRANSFORMATIONV2']._serialized_end=306
  _globals['_SUBSTRAITTRANSFORMATIONV2']._serialized_start=308
  _globals['_SUBSTRAITTRANSFORMATIONV2']._serialized_end=382
# @@protoc_insertion_point(module_scope)



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob
import os
import pathlib
import re
import shutil
import subprocess
from subprocess import CalledProcessError
import sys
from pathlib import Path

from setuptools import find_packages, setup, Command

NAME = "feast"
DESCRIPTION = "Python SDK for Feast"
URL = "https://github.com/feast-dev/feast"
AUTHOR = "Feast"
REQUIRES_PYTHON = ">=3.10.0"

REQUIRED = [
    "click>=7.0.0,<9.0.0",
    "colorama>=0.3.9,<1",
    "dill~=0.3.0",
    "protobuf>=4.24.0",
    "Jinja2>=2,<4",
    "jsonschema",
    "mmh3",
    "numpy>=2.0.0,<3",
    "pandas>=1.4.3,<3",
    "pyarrow<18.1.0",
    "pydantic>=2.0.0",
    "pygments>=2.12.0,<3",
    "PyYAML>=5.4.0,<7",
    "requests",
    "SQLAlchemy[mypy]>1",
    "tabulate>=0.8.0,<1",
    "tenacity>=7,<9",
    "toml>=0.10.0,<1",
    "tqdm>=4,<5",
    "typeguard>=4.0.0",
    "fastapi>=0.68.0",
    "uvicorn[standard]>=0.30.6,<=0.34.0",
    "uvicorn-worker",
    "gunicorn; platform_system != 'Windows'",
    "dask[dataframe]>=2024.2.1",
    "prometheus_client",
    "psutil",
    "bigtree>=0.19.2",
    "pyjwt",
]

GCP_REQUIRED = [
    "google-api-core>=1.23.0,<3",
    "googleapis-common-protos>=1.52.0,<2",
    "google-cloud-bigquery[pandas]>=2,<4",
    "google-cloud-bigquery-storage >= 2.0.0,<3",
    "google-cloud-datastore>=2.16.0,<3",
    "google-cloud-storage>=1.34.0,<3",
    "google-cloud-bigtable>=2.11.0,<3",
    "fsspec<=2024.9.0",
]

REDIS_REQUIRED = [
    "redis>=4.2.2,<5",
    "hiredis>=2.0.0,<3",
]

AWS_REQUIRED = ["boto3==1.38.27", "fsspec<=2024.9.0", "aiobotocore>2,<3"]

KUBERNETES_REQUIRED = ["kubernetes"]

SNOWFLAKE_REQUIRED = [
    "snowflake-connector-python[pandas]>=3.7,<4",
]

SPARK_REQUIRED = [
    "pyspark>=4.0.0",
]

SQLITE_VEC_REQUIRED = [
    "sqlite-vec==v0.1.6",
]
TRINO_REQUIRED = ["trino>=0.305.0,<0.400.0", "regex"]

POSTGRES_REQUIRED = [
    "psycopg[binary,pool]>=3.0.0,<4",
]
POSTGRES_C_REQUIRED = [
    "psycopg[c,pool]>=3.0.0,<4",
]

OPENTELEMETRY = ["prometheus_client", "psutil"]

MYSQL_REQUIRED = ["pymysql", "types-PyMySQL"]

HBASE_REQUIRED = [
    "happybase>=1.2.0,<3",
]

CASSANDRA_REQUIRED = [
    "cassandra-driver>=3.24.0,<4",
]

GE_REQUIRED = ["great_expectations>=0.15.41,<1"]

AZURE_REQUIRED = [
    "azure-storage-blob>=0.37.0",
    "azure-identity>=1.6.1",
    "SQLAlchemy>=1.4.19",
    "pyodbc>=4.0.30",
    "pymssql",
]

IKV_REQUIRED = [
    "ikvpy>=0.0.36",
]

HAZELCAST_REQUIRED = [
    "hazelcast-python-client>=5.1",
]

IBIS_REQUIRED = [
    "ibis-framework>=9.0.0,<10",
    "ibis-substrait>=4.0.0",
]

GRPCIO_REQUIRED = [
    "grpcio>=1.56.2,<=1.62.3",
    "grpcio-reflection>=1.56.2,<=1.62.3",
    "grpcio-health-checking>=1.56.2,<=1.62.3",
]

DUCKDB_REQUIRED = ["ibis-framework[duckdb]>=9.0.0,<10"]

DELTA_REQUIRED = ["deltalake<1.0.0"]

DOCLING_REQUIRED = ["docling>=2.23.0"]

ELASTICSEARCH_REQUIRED = ["elasticsearch>=8.13.0"]

SINGLESTORE_REQUIRED = ["singlestoredb<1.8.0"]

COUCHBASE_REQUIRED = [
    "couchbase==4.3.2",
    "couchbase-columnar==1.0.0"
]

MSSQL_REQUIRED = ["ibis-framework[mssql]>=9.0.0,<10"]

FAISS_REQUIRED = ["faiss-cpu>=1.7.0,<=1.10.0"]
QDRANT_REQUIRED = ["qdrant-client>=1.12.0"]

GO_REQUIRED = ["cffi>=1.15.0"]

MILVUS_REQUIRED = ["pymilvus"]

TORCH_REQUIRED = [
    "torch>=2.7.0",
    "torchvision>=0.22.1",
]

CLICKHOUSE_REQUIRED = ["clickhouse-connect>=0.7.19"]

MCP_REQUIRED = ["fastapi_mcp"]

RAG_REQUIRED = [
    "transformers>=4.36.0",
    "datasets>=3.6.0",
]

IMAGE_REQUIRED = [
    "timm>=0.6.0",
    "Pillow>=8.0.0",
    "scikit-learn>=1.0.0",
] + TORCH_REQUIRED

RAY_REQUIRED = [
    "ray>=2.47.0; python_version == '3.10'",
    'codeflare-sdk>=0.31.1; python_version != "3.10"',
    ]

CI_REQUIRED = (
    [
        "build",
        "virtualenv==20.23.0",
        "cryptography>=43.0,<44",
        "ruff>=0.8.0",
        "mypy-protobuf>=3.1",
        "grpcio-tools>=1.56.2,<=1.62.3",
        "grpcio-testing>=1.56.2,<=1.62.3",
        # FastAPI does not correctly pull starlette dependency on httpx see thread(https://github.com/tiangolo/fastapi/issues/5656).
        "httpx==0.27.2",
        "minio==7.2.11",
        "mock==2.0.0",
        "moto==4.2.14",
        "mypy>=1.4.1,<1.11.3",
        "urllib3>=1.25.4,<3",
        "psutil==5.9.0",
        "py>=1.11.0",  # https://github.com/pytest-dev/pytest/issues/10420
        "pytest>=6.0.0,<8",
        "pytest-asyncio<=0.24.0",
        "pytest-cov",
        "pytest-xdist",
        "pytest-benchmark>=3.4.1,<4",
        "pytest-lazy-fixture==0.6.3",
        "pytest-timeout==1.4.2",
        "pytest-ordering~=0.6.0",
        "pytest-mock==1.10.4",
        "pytest-env",
        "Sphinx>4.0.0,<7",
        "testcontainers==4.9.0",
        "python-keycloak==4.2.2",
        "pre-commit<3.3.2",
        "assertpy==1.1",
        "pip-tools",
        "pybindgen",
        "types-protobuf~=3.19.22",
        "types-python-dateutil",
        "types-pytz",
        "types-PyYAML",
        "types-redis",
        "types-requests<2.31.0",
        "types-setuptools",
        "types-tabulate",
        "virtualenv<20.24.2",
    ]
    + GCP_REQUIRED
    + REDIS_REQUIRED
    + AWS_REQUIRED
    + KUBERNETES_REQUIRED
    + SNOWFLAKE_REQUIRED
    + SPARK_REQUIRED
    + POSTGRES_REQUIRED
    + MYSQL_REQUIRED
    + TRINO_REQUIRED
    + GE_REQUIRED
    + HBASE_REQUIRED
    + CASSANDRA_REQUIRED
    + AZURE_REQUIRED
    + HAZELCAST_REQUIRED
    + IBIS_REQUIRED
    + GRPCIO_REQUIRED
    + DUCKDB_REQUIRED
    + DELTA_REQUIRED
    + ELASTICSEARCH_REQUIRED
    + SQLITE_VEC_REQUIRED
    + SINGLESTORE_REQUIRED
    + COUCHBASE_REQUIRED
    + OPENTELEMETRY
    + FAISS_REQUIRED
    + QDRANT_REQUIRED
    + MILVUS_REQUIRED
    + DOCLING_REQUIRED
    + TORCH_REQUIRED
    + CLICKHOUSE_REQUIRED
    + MCP_REQUIRED
    + RAG_REQUIRED
    + IMAGE_REQUIRED
    + RAY_REQUIRED
)
MINIMAL_REQUIRED = (
    GCP_REQUIRED
    + AWS_REQUIRED
    + REDIS_REQUIRED
    + KUBERNETES_REQUIRED
    + SNOWFLAKE_REQUIRED
    + POSTGRES_C_REQUIRED
    + MYSQL_REQUIRED
    + GO_REQUIRED
    + GRPCIO_REQUIRED
    + DUCKDB_REQUIRED
    + OPENTELEMETRY
    + MILVUS_REQUIRED
)
NLP_REQUIRED = (
    DOCLING_REQUIRED
    + MILVUS_REQUIRED
    + TORCH_REQUIRED
    + RAG_REQUIRED
    + IMAGE_REQUIRED
)
DOCS_REQUIRED = CI_REQUIRED
DEV_REQUIRED = CI_REQUIRED

# Get git repo root directory
repo_root = str(pathlib.Path(__file__).resolve().parent)

# README file from Feast repo root directory
README_FILE = os.path.join(repo_root, "README.md")
with open(README_FILE, "r", encoding="utf8") as f:
    LONG_DESCRIPTION = f.read()

# Add Support for parsing tags that have a prefix containing '/' (ie 'sdk/go') to setuptools_scm.
# Regex modified from default tag regex in:
# https://github.com/pypa/setuptools_scm/blob/2a1b46d38fb2b8aeac09853e660bcd0d7c1bc7be/src/setuptools_scm/config.py#L9
TAG_REGEX = re.compile(
    r"^(?:[\/\w-]+)?(?P<version>[vV]?\d+(?:\.\d+){0,2}[^\+]*)(?:\+.*)?$"
)

# Only set use_scm_version if git executable exists (setting this variable causes pip to use git under the hood)
if shutil.which("git"):
    use_scm_version = {"root": ".", "relative_to": __file__, "tag_regex": TAG_REGEX}
else:
    use_scm_version = None

PYTHON_CODE_PREFIX = "sdk/python"


setup(
    name=NAME,
    author=AUTHOR,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(
        where=PYTHON_CODE_PREFIX, exclude=("java", "infra", "sdk/python/tests", "ui")
    ),
    package_dir={"": PYTHON_CODE_PREFIX},
    install_requires=REQUIRED,
    extras_require={
        "dev": DEV_REQUIRED,
        "ci": CI_REQUIRED,
        "minimal": MINIMAL_REQUIRED,
        "gcp": GCP_REQUIRED,
        "aws": AWS_REQUIRED,
        "k8s": KUBERNETES_REQUIRED,
        "redis": REDIS_REQUIRED,
        "snowflake": SNOWFLAKE_REQUIRED,
        "spark": SPARK_REQUIRED,
        "trino": TRINO_REQUIRED,
        "postgres": POSTGRES_REQUIRED,
        "postgres-c": POSTGRES_C_REQUIRED,
        "azure": AZURE_REQUIRED,
        "mysql": MYSQL_REQUIRED,
        "mssql": MSSQL_REQUIRED,
        "ge": GE_REQUIRED,
        "hbase": HBASE_REQUIRED,
        "docs": DOCS_REQUIRED,
        "cassandra": CASSANDRA_REQUIRED,
        "hazelcast": HAZELCAST_REQUIRED,
        "grpcio": GRPCIO_REQUIRED,
        "ibis": IBIS_REQUIRED,
        "duckdb": DUCKDB_REQUIRED,
        "ikv": IKV_REQUIRED,
        "delta": DELTA_REQUIRED,
        "elasticsearch": ELASTICSEARCH_REQUIRED,
        "sqlite_vec": SQLITE_VEC_REQUIRED,
        "singlestore": SINGLESTORE_REQUIRED,
        "couchbase": COUCHBASE_REQUIRED,
        "opentelemetry": OPENTELEMETRY,
        "faiss": FAISS_REQUIRED,
        "qdrant": QDRANT_REQUIRED,
        "go": GO_REQUIRED,
        "milvus": MILVUS_REQUIRED,
        "docling": DOCLING_REQUIRED,
        "pytorch": TORCH_REQUIRED,
        "nlp": NLP_REQUIRED,
        "clickhouse": CLICKHOUSE_REQUIRED,
        "mcp": MCP_REQUIRED,
        "rag": RAG_REQUIRED,
        "image": IMAGE_REQUIRED,
        "ray": RAY_REQUIRED,
    },
    include_package_data=True,
    license="Apache",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={"console_scripts": ["feast=feast.cli.cli:cli"]},
    use_scm_version=use_scm_version,
    setup_requires=[
        "pybindgen==0.22.0",  # TODO do we need this?
        "setuptools_scm>=6.2",  # TODO do we need this?
    ],
)



from functools import wraps
from typing import List, Optional

from feast import utils
from feast.base_feature_view import BaseFeatureView
from feast.data_source import DataSource
from feast.entity import Entity
from feast.errors import (
    DataSourceObjectNotFoundException,
    EntityNotFoundException,
    FeatureServiceNotFoundException,
    FeatureViewNotFoundException,
    PermissionObjectNotFoundException,
    ProjectObjectNotFoundException,
    SavedDatasetNotFound,
    ValidationReferenceNotFound,
)
from feast.feature_service import FeatureService
from feast.feature_view import FeatureView
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.permissions.permission import Permission
from feast.project import Project
from feast.project_metadata import ProjectMetadata
from feast.protos.feast.core.Registry_pb2 import ProjectMetadata as ProjectMetadataProto
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.saved_dataset import SavedDataset, ValidationReference
from feast.stream_feature_view import StreamFeatureView


def registry_proto_cache(func):
    cache_key = None
    cache_value = None

    @wraps(func)
    def wrapper(registry_proto: RegistryProto, project: str):
        nonlocal cache_key, cache_value

        key = tuple([id(registry_proto), registry_proto.version_id, project])

        if key == cache_key:
            return cache_value
        else:
            cache_value = func(registry_proto, project)
            cache_key = key
            return cache_value

    return wrapper


def registry_proto_cache_with_tags(func):
    cache_key = None
    cache_value = None

    @wraps(func)
    def wrapper(
        registry_proto: RegistryProto,
        project: str,
        tags: Optional[dict[str, str]],
    ):
        nonlocal cache_key, cache_value

        key = tuple([id(registry_proto), registry_proto.version_id, project, tags])

        if key == cache_key:
            return cache_value
        else:
            cache_value = func(registry_proto, project, tags)
            cache_key = key
            return cache_value

    return wrapper


def registry_proto_cache_name(func):
    cache_proto_key = None
    cache = {}

    @wraps(func)
    def wrapper(registry_proto: RegistryProto, name: str, project: str):
        nonlocal cache_proto_key, cache

        proto_key = (id(registry_proto), registry_proto.version_id)

        if proto_key != cache_proto_key:
            cache = {}
            cache_proto_key = proto_key

        key = (project, name)
        if key in cache:
            return cache[key]
        else:
            value = func(registry_proto, name, project)
            cache[key] = value
            return value

    return wrapper


def get_project_metadata(
    registry_proto: Optional[RegistryProto], project: str
) -> Optional[ProjectMetadataProto]:
    if not registry_proto:
        return None
    for pm in registry_proto.project_metadata:
        if pm.project == project:
            return pm
    return None


def get_feature_service(
    registry_proto: RegistryProto, name: str, project: str
) -> FeatureService:
    for feature_service_proto in registry_proto.feature_services:
        if (
            feature_service_proto.spec.project == project
            and feature_service_proto.spec.name == name
        ):
            return FeatureService.from_proto(feature_service_proto)
    raise FeatureServiceNotFoundException(name, project=project)


@registry_proto_cache_name
def get_any_feature_view(
    registry_proto: RegistryProto, name: str, project: str
) -> BaseFeatureView:
    for feature_view_proto in registry_proto.feature_views:
        if (
            feature_view_proto.spec.name == name
            and feature_view_proto.spec.project == project
        ):
            return FeatureView.from_proto(feature_view_proto)

    for feature_view_proto in registry_proto.stream_feature_views:
        if (
            feature_view_proto.spec.name == name
            and feature_view_proto.spec.project == project
        ):
            return StreamFeatureView.from_proto(feature_view_proto)

    for on_demand_feature_view in registry_proto.on_demand_feature_views:
        if (
            on_demand_feature_view.spec.project == project
            and on_demand_feature_view.spec.name == name
        ):
            return OnDemandFeatureView.from_proto(on_demand_feature_view)

    raise FeatureViewNotFoundException(name, project)


def get_feature_view(
    registry_proto: RegistryProto, name: str, project: str
) -> FeatureView:
    for feature_view_proto in registry_proto.feature_views:
        if (
            feature_view_proto.spec.name == name
            and feature_view_proto.spec.project == project
        ):
            return FeatureView.from_proto(feature_view_proto)
    raise FeatureViewNotFoundException(name, project)


def get_stream_feature_view(
    registry_proto: RegistryProto, name: str, project: str
) -> StreamFeatureView:
    for feature_view_proto in registry_proto.stream_feature_views:
        if (
            feature_view_proto.spec.name == name
            and feature_view_proto.spec.project == project
        ):
            return StreamFeatureView.from_proto(feature_view_proto)
    raise FeatureViewNotFoundException(name, project)


def get_on_demand_feature_view(
    registry_proto: RegistryProto, name: str, project: str
) -> OnDemandFeatureView:
    for on_demand_feature_view in registry_proto.on_demand_feature_views:
        if (
            on_demand_feature_view.spec.project == project
            and on_demand_feature_view.spec.name == name
        ):
            return OnDemandFeatureView.from_proto(on_demand_feature_view)
    raise FeatureViewNotFoundException(name, project=project)


def get_data_source(
    registry_proto: RegistryProto, name: str, project: str
) -> DataSource:
    for data_source in registry_proto.data_sources:
        if data_source.project == project and data_source.name == name:
            return DataSource.from_proto(data_source)
    raise DataSourceObjectNotFoundException(name, project=project)


def get_entity(registry_proto: RegistryProto, name: str, project: str) -> Entity:
    for entity_proto in registry_proto.entities:
        if entity_proto.spec.name == name and entity_proto.spec.project == project:
            return Entity.from_proto(entity_proto)
    raise EntityNotFoundException(name, project=project)


def get_saved_dataset(
    registry_proto: RegistryProto, name: str, project: str
) -> SavedDataset:
    for saved_dataset in registry_proto.saved_datasets:
        if saved_dataset.spec.name == name and saved_dataset.spec.project == project:
            return SavedDataset.from_proto(saved_dataset)
    raise SavedDatasetNotFound(name, project=project)


def get_validation_reference(
    registry_proto: RegistryProto, name: str, project: str
) -> ValidationReference:
    for validation_reference in registry_proto.validation_references:
        if (
            validation_reference.name == name
            and validation_reference.project == project
        ):
            return ValidationReference.from_proto(validation_reference)
    raise ValidationReferenceNotFound(name, project=project)


@registry_proto_cache_with_tags
def list_feature_services(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[FeatureService]:
    feature_services = []
    for feature_service_proto in registry_proto.feature_services:
        if feature_service_proto.spec.project == project and utils.has_all_tags(
            feature_service_proto.spec.tags, tags
        ):
            feature_services.append(FeatureService.from_proto(feature_service_proto))
    return feature_services


@registry_proto_cache_with_tags
def list_all_feature_views(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[BaseFeatureView]:
    return (
        list_feature_views(registry_proto, project, tags)
        + list_stream_feature_views(registry_proto, project, tags)
        + list_on_demand_feature_views(registry_proto, project, tags)
    )


@registry_proto_cache_with_tags
def list_feature_views(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[FeatureView]:
    feature_views: List[FeatureView] = []
    for feature_view_proto in registry_proto.feature_views:
        if feature_view_proto.spec.project == project and utils.has_all_tags(
            feature_view_proto.spec.tags, tags
        ):
            feature_views.append(FeatureView.from_proto(feature_view_proto))
    return feature_views


@registry_proto_cache_with_tags
def list_stream_feature_views(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[StreamFeatureView]:
    stream_feature_views = []
    for stream_feature_view in registry_proto.stream_feature_views:
        if stream_feature_view.spec.project == project and utils.has_all_tags(
            stream_feature_view.spec.tags, tags
        ):
            stream_feature_views.append(
                StreamFeatureView.from_proto(stream_feature_view)
            )
    return stream_feature_views


@registry_proto_cache_with_tags
def list_on_demand_feature_views(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[OnDemandFeatureView]:
    on_demand_feature_views = []
    for on_demand_feature_view in registry_proto.on_demand_feature_views:
        if on_demand_feature_view.spec.project == project and utils.has_all_tags(
            on_demand_feature_view.spec.tags, tags
        ):
            on_demand_feature_views.append(
                OnDemandFeatureView.from_proto(on_demand_feature_view)
            )
    return on_demand_feature_views


@registry_proto_cache_with_tags
def list_entities(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[Entity]:
    entities = []
    for entity_proto in registry_proto.entities:
        if entity_proto.spec.project == project and utils.has_all_tags(
            entity_proto.spec.tags, tags
        ):
            entities.append(Entity.from_proto(entity_proto))
    return entities


@registry_proto_cache_with_tags
def list_data_sources(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[DataSource]:
    data_sources = []
    for data_source_proto in registry_proto.data_sources:
        if data_source_proto.project == project and utils.has_all_tags(
            data_source_proto.tags, tags
        ):
            data_sources.append(DataSource.from_proto(data_source_proto))
    return data_sources


@registry_proto_cache_with_tags
def list_saved_datasets(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[SavedDataset]:
    saved_datasets = []
    for saved_dataset in registry_proto.saved_datasets:
        if saved_dataset.spec.project == project and utils.has_all_tags(
            saved_dataset.spec.tags, tags
        ):
            saved_datasets.append(SavedDataset.from_proto(saved_dataset))
    return saved_datasets


@registry_proto_cache_with_tags
def list_validation_references(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[ValidationReference]:
    validation_references = []
    for validation_reference in registry_proto.validation_references:
        if validation_reference.project == project and utils.has_all_tags(
            validation_reference.tags, tags
        ):
            validation_references.append(
                ValidationReference.from_proto(validation_reference)
            )
    return validation_references


@registry_proto_cache
def list_project_metadata(
    registry_proto: RegistryProto, project: str
) -> List[ProjectMetadata]:
    return [
        ProjectMetadata.from_proto(project_metadata)
        for project_metadata in registry_proto.project_metadata
        if project_metadata.project == project
    ]


@registry_proto_cache_with_tags
def list_permissions(
    registry_proto: RegistryProto, project: str, tags: Optional[dict[str, str]]
) -> List[Permission]:
    permissions = []
    for permission_proto in registry_proto.permissions:
        if permission_proto.spec.project == project and utils.has_all_tags(
            permission_proto.spec.tags, tags
        ):
            permissions.append(Permission.from_proto(permission_proto))
    return permissions


def get_permission(
    registry_proto: RegistryProto, name: str, project: str
) -> Permission:
    for permission_proto in registry_proto.permissions:
        if (
            permission_proto.spec.project == project
            and permission_proto.spec.name == name
        ):
            return Permission.from_proto(permission_proto)
    raise PermissionObjectNotFoundException(name=name, project=project)


def list_projects(
    registry_proto: RegistryProto,
    tags: Optional[dict[str, str]],
) -> List[Project]:
    projects = []
    for project_proto in registry_proto.projects:
        if utils.has_all_tags(project_proto.spec.tags, tags):
            projects.append(Project.from_proto(project_proto))
    return projects


def get_project(registry_proto: RegistryProto, name: str) -> Project:
    for projects_proto in registry_proto.projects:
        if projects_proto.spec.name == name:
            return Project.from_proto(projects_proto)
    raise ProjectObjectNotFoundException(name=name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/Permission.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from feast.protos.feast.core import Policy_pb2 as feast_dot_core_dot_Policy__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x65\x61st/core/Permission.proto\x12\nfeast.core\x1a\x17\x66\x65\x61st/core/Policy.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"`\n\nPermission\x12(\n\x04spec\x18\x01 \x01(\x0b\x32\x1a.feast.core.PermissionSpec\x12(\n\x04meta\x18\x02 \x01(\x0b\x32\x1a.feast.core.PermissionMeta\"\xa0\x06\n\x0ePermissionSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12.\n\x05types\x18\x03 \x03(\x0e\x32\x1f.feast.core.PermissionSpec.Type\x12\x15\n\rname_patterns\x18\x04 \x03(\t\x12\x43\n\rrequired_tags\x18\x05 \x03(\x0b\x32,.feast.core.PermissionSpec.RequiredTagsEntry\x12\x39\n\x07\x61\x63tions\x18\x06 \x03(\x0e\x32(.feast.core.PermissionSpec.AuthzedAction\x12\"\n\x06policy\x18\x07 \x01(\x0b\x32\x12.feast.core.Policy\x12\x32\n\x04tags\x18\x08 \x03(\x0b\x32$.feast.core.PermissionSpec.TagsEntry\x1a\x33\n\x11RequiredTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x89\x01\n\rAuthzedAction\x12\n\n\x06\x43REATE\x10\x00\x12\x0c\n\x08\x44\x45SCRIBE\x10\x01\x12\n\n\x06UPDATE\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03\x12\x0f\n\x0bREAD_ONLINE\x10\x04\x12\x10\n\x0cREAD_OFFLINE\x10\x05\x12\x10\n\x0cWRITE_ONLINE\x10\x06\x12\x11\n\rWRITE_OFFLINE\x10\x07\"\xe1\x01\n\x04Type\x12\x10\n\x0c\x46\x45\x41TURE_VIEW\x10\x00\x12\x1a\n\x16ON_DEMAND_FEATURE_VIEW\x10\x01\x12\x16\n\x12\x42\x41TCH_FEATURE_VIEW\x10\x02\x12\x17\n\x13STREAM_FEATURE_VIEW\x10\x03\x12\n\n\x06\x45NTITY\x10\x04\x12\x13\n\x0f\x46\x45\x41TURE_SERVICE\x10\x05\x12\x0f\n\x0b\x44\x41TA_SOURCE\x10\x06\x12\x18\n\x14VALIDATION_REFERENCE\x10\x07\x12\x11\n\rSAVED_DATASET\x10\x08\x12\x0e\n\nPERMISSION\x10\t\x12\x0b\n\x07PROJECT\x10\n\"\x83\x01\n\x0ePermissionMeta\x12\x35\n\x11\x63reated_timestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampBT\n\x10\x66\x65\x61st.proto.coreB\x0fPermissionProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.Permission_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\017PermissionProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_PERMISSIONSPEC_REQUIREDTAGSENTRY']._options = None
  _globals['_PERMISSIONSPEC_REQUIREDTAGSENTRY']._serialized_options = b'8\001'
  _globals['_PERMISSIONSPEC_TAGSENTRY']._options = None
  _globals['_PERMISSIONSPEC_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_PERMISSION']._serialized_start=101
  _globals['_PERMISSION']._serialized_end=197
  _globals['_PERMISSIONSPEC']._serialized_start=200
  _globals['_PERMISSIONSPEC']._serialized_end=1000
  _globals['_PERMISSIONSPEC_REQUIREDTAGSENTRY']._serialized_start=536
  _globals['_PERMISSIONSPEC_REQUIREDTAGSENTRY']._serialized_end=587
  _globals['_PERMISSIONSPEC_TAGSENTRY']._serialized_start=589
  _globals['_PERMISSIONSPEC_TAGSENTRY']._serialized_end=632
  _globals['_PERMISSIONSPEC_AUTHZEDACTION']._serialized_start=635
  _globals['_PERMISSIONSPEC_AUTHZEDACTION']._serialized_end=772
  _globals['_PERMISSIONSPEC_TYPE']._serialized_start=775
  _globals['_PERMISSIONSPEC_TYPE']._serialized_end=1000
  _globals['_PERMISSIONMETA']._serialized_start=1003
  _globals['_PERMISSIONMETA']._serialized_end=1134
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Policy(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    PROJECT_FIELD_NUMBER: builtins.int
    ROLE_BASED_POLICY_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Name of the policy."""
    project: builtins.str
    """Name of Feast project."""
    @property
    def role_based_policy(self) -> global___RoleBasedPolicy: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        project: builtins.str = ...,
        role_based_policy: global___RoleBasedPolicy | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["policy_type", b"policy_type", "role_based_policy", b"role_based_policy"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name", "policy_type", b"policy_type", "project", b"project", "role_based_policy", b"role_based_policy"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["policy_type", b"policy_type"]) -> typing_extensions.Literal["role_based_policy"] | None: ...

global___Policy = Policy

class RoleBasedPolicy(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ROLES_FIELD_NUMBER: builtins.int
    @property
    def roles(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """List of roles in this policy."""
    def __init__(
        self,
        *,
        roles: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["roles", b"roles"]) -> None: ...

global___RoleBasedPolicy = RoleBasedPolicy



from typing import TYPE_CHECKING, Dict, List, Optional

from attr import dataclass

from feast.data_source import DataSource
from feast.field import Field
from feast.protos.feast.core.FeatureViewProjection_pb2 import (
    FeatureViewProjection as FeatureViewProjectionProto,
)

if TYPE_CHECKING:
    from feast.base_feature_view import BaseFeatureView
    from feast.feature_view import FeatureView


@dataclass
class FeatureViewProjection:
    """
    A feature view projection represents a selection of one or more features from a
    single feature view.

    Attributes:
        name: The unique name of the feature view from which this projection is created.
        name_alias: An optional alias for the name.
        features: The list of features represented by the feature view projection.
        desired_features: The list of features that this feature view projection intends to select.
            If empty, the projection intends to select all features. This attribute is only used
            for feature service inference. It should only be set if the underlying feature view
            is not ready to be projected, i.e. still needs to go through feature inference.
        join_key_map: A map to modify join key columns during retrieval of this feature
            view projection.
        timestamp_field: The timestamp field of the feature view projection.
        date_partition_column: The date partition column of the feature view projection.
        created_timestamp_column: The created timestamp column of the feature view projection.
        batch_source: The batch source of data where this group of features
            is stored. This is optional ONLY if a push source is specified as the
            stream_source, since push sources contain their own batch sources.

    """

    name: str
    name_alias: Optional[str]
    desired_features: List[str]
    features: List[Field]
    join_key_map: Dict[str, str] = {}
    timestamp_field: Optional[str] = None
    date_partition_column: Optional[str] = None
    created_timestamp_column: Optional[str] = None
    batch_source: Optional[DataSource] = None

    def name_to_use(self):
        return self.name_alias or self.name

    def to_proto(self) -> FeatureViewProjectionProto:
        batch_source = None
        if getattr(self, "batch_source", None):
            if isinstance(self.batch_source, DataSource):
                batch_source = self.batch_source.to_proto()
            else:
                batch_source = self.batch_source
        feature_reference_proto = FeatureViewProjectionProto(
            feature_view_name=self.name,
            feature_view_name_alias=self.name_alias or "",
            join_key_map=self.join_key_map,
            timestamp_field=self.timestamp_field or "",
            date_partition_column=self.date_partition_column or "",
            created_timestamp_column=self.created_timestamp_column or "",
            batch_source=batch_source,
        )
        for feature in self.features:
            feature_reference_proto.feature_columns.append(feature.to_proto())

        return feature_reference_proto

    @staticmethod
    def from_proto(proto: FeatureViewProjectionProto) -> "FeatureViewProjection":
        batch_source = (
            DataSource.from_proto(proto.batch_source)
            if str(getattr(proto, "batch_source"))
            else None
        )
        feature_view_projection = FeatureViewProjection(
            name=proto.feature_view_name,
            name_alias=proto.feature_view_name_alias or None,
            features=[],
            join_key_map=dict(proto.join_key_map),
            desired_features=[],
            timestamp_field=proto.timestamp_field or None,
            date_partition_column=proto.date_partition_column or None,
            created_timestamp_column=proto.created_timestamp_column or None,
            batch_source=batch_source,
        )
        for feature_column in proto.feature_columns:
            feature_view_projection.features.append(Field.from_proto(feature_column))

        return feature_view_projection

    @staticmethod
    def from_feature_view_definition(feature_view: "FeatureView"):
        # TODO need to implement this for StreamFeatureViews
        if getattr(feature_view, "batch_source", None):
            return FeatureViewProjection(
                name=feature_view.name,
                name_alias=None,
                features=feature_view.features,
                desired_features=[],
                timestamp_field=feature_view.batch_source.created_timestamp_column
                or None,
                created_timestamp_column=feature_view.batch_source.created_timestamp_column
                or None,
                date_partition_column=feature_view.batch_source.date_partition_column
                or None,
                batch_source=feature_view.batch_source or None,
            )
        else:
            return FeatureViewProjection(
                name=feature_view.name,
                name_alias=None,
                features=feature_view.features,
                desired_features=[],
            )

    @staticmethod
    def from_definition(base_feature_view: "BaseFeatureView"):
        if getattr(base_feature_view, "batch_source", None):
            return FeatureViewProjection(
                name=base_feature_view.name,
                name_alias=None,
                features=base_feature_view.features,
                desired_features=[],
                timestamp_field=base_feature_view.batch_source.created_timestamp_column  # type:ignore[attr-defined]
                or None,
                created_timestamp_column=base_feature_view.batch_source.created_timestamp_column  # type:ignore[attr-defined]
                or None,
                date_partition_column=base_feature_view.batch_source.date_partition_column  # type:ignore[attr-defined]
                or None,
                batch_source=base_feature_view.batch_source or None,  # type:ignore[attr-defined]
            )
        else:
            return FeatureViewProjection(
                name=base_feature_view.name,
                name_alias=None,
                features=base_feature_view.features,
                desired_features=[],
            )

    def get_feature(self, feature_name: str) -> Field:
        try:
            return next(field for field in self.features if field.name == feature_name)
        except StopIteration:
            raise KeyError(
                f"Feature {feature_name} not found in projection {self.name_to_use()}"
            )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Copyright 2021 The Feast Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _TransformationServiceType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _TransformationServiceTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_TransformationServiceType.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    TRANSFORMATION_SERVICE_TYPE_INVALID: _TransformationServiceType.ValueType  # 0
    TRANSFORMATION_SERVICE_TYPE_PYTHON: _TransformationServiceType.ValueType  # 1
    TRANSFORMATION_SERVICE_TYPE_CUSTOM: _TransformationServiceType.ValueType  # 100

class TransformationServiceType(_TransformationServiceType, metaclass=_TransformationServiceTypeEnumTypeWrapper): ...

TRANSFORMATION_SERVICE_TYPE_INVALID: TransformationServiceType.ValueType  # 0
TRANSFORMATION_SERVICE_TYPE_PYTHON: TransformationServiceType.ValueType  # 1
TRANSFORMATION_SERVICE_TYPE_CUSTOM: TransformationServiceType.ValueType  # 100
global___TransformationServiceType = TransformationServiceType

class ValueType(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ARROW_VALUE_FIELD_NUMBER: builtins.int
    arrow_value: builtins.bytes
    """Having a oneOf provides forward compatibility if we need to support compound types
    that are not supported by arrow natively.
    """
    def __init__(
        self,
        *,
        arrow_value: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["arrow_value", b"arrow_value", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["arrow_value", b"arrow_value", "value", b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["value", b"value"]) -> typing_extensions.Literal["arrow_value"] | None: ...

global___ValueType = ValueType

class GetTransformationServiceInfoRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetTransformationServiceInfoRequest = GetTransformationServiceInfoRequest

class GetTransformationServiceInfoResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VERSION_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    TRANSFORMATION_SERVICE_TYPE_DETAILS_FIELD_NUMBER: builtins.int
    version: builtins.str
    """Feast version of this transformation service deployment."""
    type: global___TransformationServiceType.ValueType
    """Type of transformation service deployment. This is either Python, or custom"""
    transformation_service_type_details: builtins.str
    def __init__(
        self,
        *,
        version: builtins.str = ...,
        type: global___TransformationServiceType.ValueType = ...,
        transformation_service_type_details: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["transformation_service_type_details", b"transformation_service_type_details", "type", b"type", "version", b"version"]) -> None: ...

global___GetTransformationServiceInfoResponse = GetTransformationServiceInfoResponse

class TransformFeaturesRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ON_DEMAND_FEATURE_VIEW_NAME_FIELD_NUMBER: builtins.int
    PROJECT_FIELD_NUMBER: builtins.int
    TRANSFORMATION_INPUT_FIELD_NUMBER: builtins.int
    on_demand_feature_view_name: builtins.str
    project: builtins.str
    @property
    def transformation_input(self) -> global___ValueType: ...
    def __init__(
        self,
        *,
        on_demand_feature_view_name: builtins.str = ...,
        project: builtins.str = ...,
        transformation_input: global___ValueType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["transformation_input", b"transformation_input"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["on_demand_feature_view_name", b"on_demand_feature_view_name", "project", b"project", "transformation_input", b"transformation_input"]) -> None: ...

global___TransformFeaturesRequest = TransformFeaturesRequest

class TransformFeaturesResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRANSFORMATION_OUTPUT_FIELD_NUMBER: builtins.int
    @property
    def transformation_output(self) -> global___ValueType: ...
    def __init__(
        self,
        *,
        transformation_output: global___ValueType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["transformation_output", b"transformation_output"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["transformation_output", b"transformation_output"]) -> None: ...

global___TransformFeaturesResponse = TransformFeaturesResponse



# # # # # # # # # # # # # # # # # # # # # # # #
# This is an example feature definition file  #
# showcasing Ray offline store and compute     #
# engine capabilities                          #
# # # # # # # # # # # # # # # # # # # # # # # #

from datetime import timedelta
from pathlib import Path

from feast import Entity, FeatureService, FeatureView, Field, ValueType
from feast.infra.offline_stores.file_source import FileSource
from feast.on_demand_feature_view import on_demand_feature_view
from feast.types import Float32, Float64, Int64

# Constants related to the generated data sets
CURRENT_DIR = Path(__file__).parent

# Entity definitions
driver = Entity(
    name="driver",
    description="driver id",
    value_type=ValueType.INT64,
    join_keys=["driver_id"],
)

customer = Entity(
    name="customer",
    description="customer id",
    value_type=ValueType.INT64,
    join_keys=["customer_id"],
)

# Data sources - Ray offline store works with FileSource
# These will be processed by Ray for efficient distributed data access
driver_hourly_stats = FileSource(
    name="driver_hourly_stats",
    path=f"{CURRENT_DIR}/%PARQUET_PATH%",
    timestamp_field="event_timestamp",
    created_timestamp_column="created",
)

customer_daily_profile = FileSource(
    name="customer_daily_profile",
    path=f"{CURRENT_DIR}/data/customer_daily_profile.parquet",
    timestamp_field="event_timestamp",
    created_timestamp_column="created",
)

# Feature Views - These leverage Ray compute engine for distributed processing
driver_hourly_stats_view = FeatureView(
    name="driver_hourly_stats",
    entities=[driver],
    ttl=timedelta(days=7),
    schema=[
        Field(name="conv_rate", dtype=Float32),
        Field(name="acc_rate", dtype=Float32),
        Field(name="avg_daily_trips", dtype=Int64),
    ],
    online=True,
    source=driver_hourly_stats,
    tags={"team": "driver_performance", "processing": "ray"},
)

customer_daily_profile_view = FeatureView(
    name="customer_daily_profile",
    entities=[customer],
    ttl=timedelta(days=7),
    schema=[
        Field(name="current_balance", dtype=Float32),
        Field(name="avg_passenger_count", dtype=Float32),
        Field(name="lifetime_trip_count", dtype=Int64),
    ],
    online=True,
    source=customer_daily_profile,
    tags={"team": "customer_analytics", "processing": "ray"},
)


# On-demand feature view showcasing Ray compute engine capabilities
# This demonstrates real-time feature transformations using Ray
@on_demand_feature_view(
    sources=[driver_hourly_stats_view],
    schema=[
        Field(name="conv_rate_plus_acc_rate", dtype=Float64),
        Field(name="trips_per_day_normalized", dtype=Float64),
    ],
)
def driver_activity_v2(inputs: dict):
    """
    On-demand feature transformations processed by Ray compute engine.
    These calculations happen in real-time and can leverage Ray's
    distributed processing capabilities.
    """
    import pandas as pd

    conv_rate = inputs["conv_rate"]
    acc_rate = inputs["acc_rate"]
    avg_daily_trips = inputs["avg_daily_trips"]

    # Feature engineering using Ray's distributed processing
    conv_rate_plus_acc_rate = conv_rate + acc_rate

    # Normalize trips per day (example of more complex transformation)
    max_trips = avg_daily_trips.max() if len(avg_daily_trips) > 0 else 1
    trips_per_day_normalized = avg_daily_trips / max_trips

    return pd.DataFrame(
        {
            "conv_rate_plus_acc_rate": conv_rate_plus_acc_rate,
            "trips_per_day_normalized": trips_per_day_normalized,
        }
    )


# Feature Service - Groups related features for serving
# Ray compute engine optimizes the retrieval of these feature combinations
driver_activity_v1 = FeatureService(
    name="driver_activity_v1",
    features=[
        driver_hourly_stats_view,
        customer_daily_profile_view,
    ],
    tags={"version": "v1", "compute_engine": "ray"},
)

driver_activity_v2_service = FeatureService(
    name="driver_activity_v2",
    features=[
        driver_hourly_stats_view,
        customer_daily_profile_view,
        driver_activity_v2,  # Includes on-demand transformations
    ],
    tags={"version": "v2", "compute_engine": "ray", "transformations": "on_demand"},
)



# Registry lineage generation utilities
from .registry_lineage import EntityReference, EntityRelation, RegistryLineageGenerator

__all__ = ["RegistryLineageGenerator", "EntityRelation", "EntityReference"]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

* Copyright 2019 The Feast Authors
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     https://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Store(google.protobuf.message.Message):
    """Store provides a location where Feast reads and writes feature values.
    Feature values will be written to the Store in the form of FeatureRow elements.
    The way FeatureRow is encoded and decoded when it is written to and read from
    the Store depends on the type of the Store.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _StoreType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _StoreTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Store._StoreType.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        INVALID: Store._StoreType.ValueType  # 0
        REDIS: Store._StoreType.ValueType  # 1
        """Redis stores a FeatureRow element as a key, value pair.

        The Redis data types used (https://redis.io/topics/data-types):
        - key: STRING
        - value: STRING

        Encodings:
        - key: byte array of RedisKey (refer to feast.storage.RedisKeyV2)
        - value: Redis hashmap
        """
        REDIS_CLUSTER: Store._StoreType.ValueType  # 4

    class StoreType(_StoreType, metaclass=_StoreTypeEnumTypeWrapper): ...
    INVALID: Store.StoreType.ValueType  # 0
    REDIS: Store.StoreType.ValueType  # 1
    """Redis stores a FeatureRow element as a key, value pair.

    The Redis data types used (https://redis.io/topics/data-types):
    - key: STRING
    - value: STRING

    Encodings:
    - key: byte array of RedisKey (refer to feast.storage.RedisKeyV2)
    - value: Redis hashmap
    """
    REDIS_CLUSTER: Store.StoreType.ValueType  # 4

    class RedisConfig(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        HOST_FIELD_NUMBER: builtins.int
        PORT_FIELD_NUMBER: builtins.int
        INITIAL_BACKOFF_MS_FIELD_NUMBER: builtins.int
        MAX_RETRIES_FIELD_NUMBER: builtins.int
        FLUSH_FREQUENCY_SECONDS_FIELD_NUMBER: builtins.int
        SSL_FIELD_NUMBER: builtins.int
        host: builtins.str
        port: builtins.int
        initial_backoff_ms: builtins.int
        """Optional. The number of milliseconds to wait before retrying failed Redis connection.
        By default, Feast uses exponential backoff policy and "initial_backoff_ms" sets the initial wait duration.
        """
        max_retries: builtins.int
        """Optional. Maximum total number of retries for connecting to Redis. Default to zero retries."""
        flush_frequency_seconds: builtins.int
        """Optional. How often flush data to redis"""
        ssl: builtins.bool
        """Optional. Connect over SSL."""
        def __init__(
            self,
            *,
            host: builtins.str = ...,
            port: builtins.int = ...,
            initial_backoff_ms: builtins.int = ...,
            max_retries: builtins.int = ...,
            flush_frequency_seconds: builtins.int = ...,
            ssl: builtins.bool = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["flush_frequency_seconds", b"flush_frequency_seconds", "host", b"host", "initial_backoff_ms", b"initial_backoff_ms", "max_retries", b"max_retries", "port", b"port", "ssl", b"ssl"]) -> None: ...

    class RedisClusterConfig(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _ReadFrom:
            ValueType = typing.NewType("ValueType", builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _ReadFromEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Store.RedisClusterConfig._ReadFrom.ValueType], builtins.type):  # noqa: F821
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            MASTER: Store.RedisClusterConfig._ReadFrom.ValueType  # 0
            MASTER_PREFERRED: Store.RedisClusterConfig._ReadFrom.ValueType  # 1
            REPLICA: Store.RedisClusterConfig._ReadFrom.ValueType  # 2
            REPLICA_PREFERRED: Store.RedisClusterConfig._ReadFrom.ValueType  # 3

        class ReadFrom(_ReadFrom, metaclass=_ReadFromEnumTypeWrapper):
            """Optional. Priority of nodes when reading from cluster"""

        MASTER: Store.RedisClusterConfig.ReadFrom.ValueType  # 0
        MASTER_PREFERRED: Store.RedisClusterConfig.ReadFrom.ValueType  # 1
        REPLICA: Store.RedisClusterConfig.ReadFrom.ValueType  # 2
        REPLICA_PREFERRED: Store.RedisClusterConfig.ReadFrom.ValueType  # 3

        CONNECTION_STRING_FIELD_NUMBER: builtins.int
        INITIAL_BACKOFF_MS_FIELD_NUMBER: builtins.int
        MAX_RETRIES_FIELD_NUMBER: builtins.int
        FLUSH_FREQUENCY_SECONDS_FIELD_NUMBER: builtins.int
        KEY_PREFIX_FIELD_NUMBER: builtins.int
        ENABLE_FALLBACK_FIELD_NUMBER: builtins.int
        FALLBACK_PREFIX_FIELD_NUMBER: builtins.int
        READ_FROM_FIELD_NUMBER: builtins.int
        connection_string: builtins.str
        """List of Redis Uri for all the nodes in Redis Cluster, comma separated. Eg. host1:6379, host2:6379"""
        initial_backoff_ms: builtins.int
        max_retries: builtins.int
        flush_frequency_seconds: builtins.int
        """Optional. How often flush data to redis"""
        key_prefix: builtins.str
        """Optional. Append a prefix to the Redis Key"""
        enable_fallback: builtins.bool
        """Optional. Enable fallback to another key prefix if the original key is not present.
        Useful for migrating key prefix without re-ingestion. Disabled by default.
        """
        fallback_prefix: builtins.str
        """Optional. This would be the fallback prefix to use if enable_fallback is true."""
        read_from: global___Store.RedisClusterConfig.ReadFrom.ValueType
        def __init__(
            self,
            *,
            connection_string: builtins.str = ...,
            initial_backoff_ms: builtins.int = ...,
            max_retries: builtins.int = ...,
            flush_frequency_seconds: builtins.int = ...,
            key_prefix: builtins.str = ...,
            enable_fallback: builtins.bool = ...,
            fallback_prefix: builtins.str = ...,
            read_from: global___Store.RedisClusterConfig.ReadFrom.ValueType = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["connection_string", b"connection_string", "enable_fallback", b"enable_fallback", "fallback_prefix", b"fallback_prefix", "flush_frequency_seconds", b"flush_frequency_seconds", "initial_backoff_ms", b"initial_backoff_ms", "key_prefix", b"key_prefix", "max_retries", b"max_retries", "read_from", b"read_from"]) -> None: ...

    class Subscription(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        PROJECT_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        EXCLUDE_FIELD_NUMBER: builtins.int
        project: builtins.str
        """Name of project that the feature sets belongs to. This can be one of
        - [project_name]
        - *
        If an asterisk is provided, filtering on projects will be disabled. All projects will
        be matched. It is NOT possible to provide an asterisk with a string in order to do
        pattern matching.
        """
        name: builtins.str
        """Name of the desired feature set. Asterisks can be used as wildcards in the name.
        Matching on names is only permitted if a specific project is defined. It is disallowed
        If the project name is set to "*"
        e.g.
        - * can be used to match all feature sets
        - my-feature-set* can be used to match all features prefixed by "my-feature-set"
        - my-feature-set-6 can be used to select a single feature set
        """
        exclude: builtins.bool
        """All matches with exclude enabled will be filtered out instead of added"""
        def __init__(
            self,
            *,
            project: builtins.str = ...,
            name: builtins.str = ...,
            exclude: builtins.bool = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["exclude", b"exclude", "name", b"name", "project", b"project"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    SUBSCRIPTIONS_FIELD_NUMBER: builtins.int
    REDIS_CONFIG_FIELD_NUMBER: builtins.int
    REDIS_CLUSTER_CONFIG_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Name of the store."""
    type: global___Store.StoreType.ValueType
    """Type of store."""
    @property
    def subscriptions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Store.Subscription]:
        """Feature sets to subscribe to."""
    @property
    def redis_config(self) -> global___Store.RedisConfig: ...
    @property
    def redis_cluster_config(self) -> global___Store.RedisClusterConfig: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        type: global___Store.StoreType.ValueType = ...,
        subscriptions: collections.abc.Iterable[global___Store.Subscription] | None = ...,
        redis_config: global___Store.RedisConfig | None = ...,
        redis_cluster_config: global___Store.RedisClusterConfig | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["config", b"config", "redis_cluster_config", b"redis_cluster_config", "redis_config", b"redis_config"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["config", b"config", "name", b"name", "redis_cluster_config", b"redis_cluster_config", "redis_config", b"redis_config", "subscriptions", b"subscriptions", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["config", b"config"]) -> typing_extensions.Literal["redis_config", "redis_cluster_config"] | None: ...

global___Store = Store



from typing import Any, Dict, Optional, Union, cast

import pandas as pd
import pyspark.sql

from feast.infra.compute_engines.spark.utils import get_or_create_new_spark_session
from feast.transformation.base import Transformation
from feast.transformation.mode import TransformationMode


class SparkTransformation(Transformation):
    r"""
    SparkTransformation can be used to define a transformation using a Spark UDF or SQL query.
    The current spark session will be used or a new one will be created if not available.
    E.g.:
    spark_transformation = SparkTransformation(
        mode=TransformationMode.SPARK,
        udf=remove_extra_spaces,
        udf_string="remove extra spaces",
    )
    OR
    spark_transformation = Transformation(
        mode=TransformationMode.SPARK_SQL,
        udf=remove_extra_spaces_sql,
        udf_string="remove extra spaces sql",
    )
    OR
    @transformation(mode=TransformationMode.SPARK)
    def remove_extra_spaces_udf(df: pd.DataFrame) -> pd.DataFrame:
        return df.assign(name=df['name'].str.replace('\s+', ' '))
    """

    def __new__(
        cls,
        mode: Union[TransformationMode, str],
        udf: Any,
        udf_string: str,
        spark_config: Dict[str, Any] = {},
        name: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        description: str = "",
        owner: str = "",
        *args,
        **kwargs,
    ) -> "SparkTransformation":
        """
        Creates a SparkTransformation
        Args:
            mode: (required) The mode of the transformation. Choose one from TransformationMode.SPARK or TransformationMode.SPARK_SQL.
            udf: (required) The user-defined transformation function.
            udf_string: (required) The string representation of the udf. The dill get source doesn't
            spark_config: (optional) The spark configuration to use for the transformation.
            name: (optional) The name of the transformation.
            tags: (optional) Metadata tags for the transformation.
            description: (optional) A description of the transformation.
            owner: (optional) The owner of the transformation.
        """
        instance = super(SparkTransformation, cls).__new__(
            cls,
            mode=mode,
            spark_config=spark_config,
            udf=udf,
            udf_string=udf_string,
            name=name,
            tags=tags,
            description=description,
            owner=owner,
        )
        return cast(SparkTransformation, instance)

    def __init__(
        self,
        mode: Union[TransformationMode, str],
        udf: Any,
        udf_string: str,
        spark_config: Dict[str, Any] = {},
        name: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        description: str = "",
        owner: str = "",
        *args,
        **kwargs,
    ):
        super().__init__(
            mode=mode,
            udf=udf,
            name=name,
            udf_string=udf_string,
            tags=tags,
            description=description,
            owner=owner,
        )
        self.spark_session = get_or_create_new_spark_session(spark_config)

    def transform(
        self,
        *inputs: Union[str, pd.DataFrame],
    ) -> pd.DataFrame:
        if self.mode == TransformationMode.SPARK_SQL:
            return self._transform_spark_sql(*inputs)
        else:
            return self._transform_spark_udf(*inputs)

    @staticmethod
    def _create_temp_view_for_dataframe(df: pyspark.sql.DataFrame, name: str):
        df_temp_view = f"feast_transformation_temp_view_{name}"
        df.createOrReplaceTempView(df_temp_view)
        return df_temp_view

    def _transform_spark_sql(
        self, *inputs: Union[pyspark.sql.DataFrame, str]
    ) -> pd.DataFrame:
        inputs_str = [
            self._create_temp_view_for_dataframe(v, f"index_{i}")
            if isinstance(v, pyspark.sql.DataFrame)
            else v
            for i, v in enumerate(inputs)
        ]
        return self.spark_session.sql(self.udf(*inputs_str))

    def _transform_spark_udf(self, *inputs: Any) -> pd.DataFrame:
        return self.udf(*inputs)

    def infer_features(self, *args, **kwargs) -> Any:
        pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/ValidationProfile.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"feast/core/ValidationProfile.proto\x12\nfeast.core\"\x83\x01\n\x14GEValidationProfiler\x12\x46\n\x08profiler\x18\x01 \x01(\x0b\x32\x34.feast.core.GEValidationProfiler.UserDefinedProfiler\x1a#\n\x13UserDefinedProfiler\x12\x0c\n\x04\x62ody\x18\x01 \x01(\x0c\"0\n\x13GEValidationProfile\x12\x19\n\x11\x65xpectation_suite\x18\x01 \x01(\x0c\"\xdd\x02\n\x13ValidationReference\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1e\n\x16reference_dataset_name\x18\x02 \x01(\t\x12\x0f\n\x07project\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x37\n\x04tags\x18\x05 \x03(\x0b\x32).feast.core.ValidationReference.TagsEntry\x12\x37\n\x0bge_profiler\x18\x06 \x01(\x0b\x32 .feast.core.GEValidationProfilerH\x00\x12\x35\n\nge_profile\x18\x07 \x01(\x0b\x32\x1f.feast.core.GEValidationProfileH\x01\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\n\n\x08profilerB\x10\n\x0e\x63\x61\x63hed_profileBV\n\x10\x66\x65\x61st.proto.coreB\x11ValidationProfileZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.ValidationProfile_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\021ValidationProfileZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_VALIDATIONREFERENCE_TAGSENTRY']._options = None
  _globals['_VALIDATIONREFERENCE_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_GEVALIDATIONPROFILER']._serialized_start=51
  _globals['_GEVALIDATIONPROFILER']._serialized_end=182
  _globals['_GEVALIDATIONPROFILER_USERDEFINEDPROFILER']._serialized_start=147
  _globals['_GEVALIDATIONPROFILER_USERDEFINEDPROFILER']._serialized_end=182
  _globals['_GEVALIDATIONPROFILE']._serialized_start=184
  _globals['_GEVALIDATIONPROFILE']._serialized_end=232
  _globals['_VALIDATIONREFERENCE']._serialized_start=235
  _globals['_VALIDATIONREFERENCE']._serialized_end=584
  _globals['_VALIDATIONREFERENCE_TAGSENTRY']._serialized_start=511
  _globals['_VALIDATIONREFERENCE_TAGSENTRY']._serialized_end=554
# @@protoc_insertion_point(module_scope)



from starlette.authentication import (
    AuthenticationError,
)

from feast.permissions.auth.token_extractor import TokenExtractor


class ArrowFlightTokenExtractor(TokenExtractor):
    def extract_access_token(self, **kwargs) -> str:
        """
        Token extractor for Arrow Flight requests.

        Requires a keyword argument called `headers` of type `dict`.

        Returns:
            The extracted access token.
        """

        if "headers" not in kwargs:
            raise ValueError("Missing keywork argument 'headers'")
        if not isinstance(kwargs["headers"], dict):
            raise ValueError(
                f"The keywork argument 'headers' is not of the expected type {dict.__name__}"
            )

        access_token = None
        headers = kwargs["headers"]
        if isinstance(headers, dict):
            for header in headers:
                if header.lower() == "authorization":
                    # With Arrow Flight, the header value is a list and we take the 0-th element
                    if not isinstance(headers[header], list):
                        raise AuthenticationError(
                            f"Authorization header must be of type list, found {type(headers[header])}"
                        )

                    return self._extract_bearer_token(headers[header][0])

        if access_token is None:
            raise AuthenticationError("Missing authorization header")

        return access_token



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import feast.core.FeatureViewProjection_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class FeatureService(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SPEC_FIELD_NUMBER: builtins.int
    META_FIELD_NUMBER: builtins.int
    @property
    def spec(self) -> global___FeatureServiceSpec:
        """User-specified specifications of this feature service."""
    @property
    def meta(self) -> global___FeatureServiceMeta:
        """System-populated metadata for this feature service."""
    def __init__(
        self,
        *,
        spec: global___FeatureServiceSpec | None = ...,
        meta: global___FeatureServiceMeta | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> None: ...

global___FeatureService = FeatureService

class FeatureServiceSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    PROJECT_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    OWNER_FIELD_NUMBER: builtins.int
    LOGGING_CONFIG_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Name of the Feature Service. Must be unique. Not updated."""
    project: builtins.str
    """Name of Feast project that this Feature Service belongs to."""
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.FeatureViewProjection_pb2.FeatureViewProjection]:
        """Represents a projection that's to be applied on top of the FeatureView.
        Contains data such as the features to use from a FeatureView.
        """
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """User defined metadata"""
    description: builtins.str
    """Description of the feature service."""
    owner: builtins.str
    """Owner of the feature service."""
    @property
    def logging_config(self) -> global___LoggingConfig:
        """(optional) if provided logging will be enabled for this feature service."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        project: builtins.str = ...,
        features: collections.abc.Iterable[feast.core.FeatureViewProjection_pb2.FeatureViewProjection] | None = ...,
        tags: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        description: builtins.str = ...,
        owner: builtins.str = ...,
        logging_config: global___LoggingConfig | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["logging_config", b"logging_config"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["description", b"description", "features", b"features", "logging_config", b"logging_config", "name", b"name", "owner", b"owner", "project", b"project", "tags", b"tags"]) -> None: ...

global___FeatureServiceSpec = FeatureServiceSpec

class FeatureServiceMeta(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    LAST_UPDATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    @property
    def created_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time where this Feature Service is created"""
    @property
    def last_updated_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time where this Feature Service is last updated"""
    def __init__(
        self,
        *,
        created_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        last_updated_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp"]) -> None: ...

global___FeatureServiceMeta = FeatureServiceMeta

class LoggingConfig(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class FileDestination(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        PATH_FIELD_NUMBER: builtins.int
        S3_ENDPOINT_OVERRIDE_FIELD_NUMBER: builtins.int
        PARTITION_BY_FIELD_NUMBER: builtins.int
        path: builtins.str
        s3_endpoint_override: builtins.str
        @property
        def partition_by(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
            """column names to use for partitioning"""
        def __init__(
            self,
            *,
            path: builtins.str = ...,
            s3_endpoint_override: builtins.str = ...,
            partition_by: collections.abc.Iterable[builtins.str] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["partition_by", b"partition_by", "path", b"path", "s3_endpoint_override", b"s3_endpoint_override"]) -> None: ...

    class BigQueryDestination(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TABLE_REF_FIELD_NUMBER: builtins.int
        table_ref: builtins.str
        """Full table reference in the form of [project:dataset.table]"""
        def __init__(
            self,
            *,
            table_ref: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["table_ref", b"table_ref"]) -> None: ...

    class RedshiftDestination(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TABLE_NAME_FIELD_NUMBER: builtins.int
        table_name: builtins.str
        """Destination table name. ClusterId and database will be taken from an offline store config"""
        def __init__(
            self,
            *,
            table_name: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["table_name", b"table_name"]) -> None: ...

    class AthenaDestination(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TABLE_NAME_FIELD_NUMBER: builtins.int
        table_name: builtins.str
        """Destination table name. data_source and database will be taken from an offline store config"""
        def __init__(
            self,
            *,
            table_name: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["table_name", b"table_name"]) -> None: ...

    class SnowflakeDestination(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TABLE_NAME_FIELD_NUMBER: builtins.int
        table_name: builtins.str
        """Destination table name. Schema and database will be taken from an offline store config"""
        def __init__(
            self,
            *,
            table_name: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["table_name", b"table_name"]) -> None: ...

    class CustomDestination(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class ConfigEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            value: builtins.str
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: builtins.str = ...,
            ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

        KIND_FIELD_NUMBER: builtins.int
        CONFIG_FIELD_NUMBER: builtins.int
        kind: builtins.str
        @property
        def config(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
        def __init__(
            self,
            *,
            kind: builtins.str = ...,
            config: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["config", b"config", "kind", b"kind"]) -> None: ...

    class CouchbaseColumnarDestination(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DATABASE_FIELD_NUMBER: builtins.int
        SCOPE_FIELD_NUMBER: builtins.int
        COLLECTION_FIELD_NUMBER: builtins.int
        database: builtins.str
        """Destination database name"""
        scope: builtins.str
        """Destination scope name"""
        collection: builtins.str
        """Destination collection name"""
        def __init__(
            self,
            *,
            database: builtins.str = ...,
            scope: builtins.str = ...,
            collection: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["collection", b"collection", "database", b"database", "scope", b"scope"]) -> None: ...

    SAMPLE_RATE_FIELD_NUMBER: builtins.int
    FILE_DESTINATION_FIELD_NUMBER: builtins.int
    BIGQUERY_DESTINATION_FIELD_NUMBER: builtins.int
    REDSHIFT_DESTINATION_FIELD_NUMBER: builtins.int
    SNOWFLAKE_DESTINATION_FIELD_NUMBER: builtins.int
    CUSTOM_DESTINATION_FIELD_NUMBER: builtins.int
    ATHENA_DESTINATION_FIELD_NUMBER: builtins.int
    COUCHBASE_COLUMNAR_DESTINATION_FIELD_NUMBER: builtins.int
    sample_rate: builtins.float
    @property
    def file_destination(self) -> global___LoggingConfig.FileDestination: ...
    @property
    def bigquery_destination(self) -> global___LoggingConfig.BigQueryDestination: ...
    @property
    def redshift_destination(self) -> global___LoggingConfig.RedshiftDestination: ...
    @property
    def snowflake_destination(self) -> global___LoggingConfig.SnowflakeDestination: ...
    @property
    def custom_destination(self) -> global___LoggingConfig.CustomDestination: ...
    @property
    def athena_destination(self) -> global___LoggingConfig.AthenaDestination: ...
    @property
    def couchbase_columnar_destination(self) -> global___LoggingConfig.CouchbaseColumnarDestination: ...
    def __init__(
        self,
        *,
        sample_rate: builtins.float = ...,
        file_destination: global___LoggingConfig.FileDestination | None = ...,
        bigquery_destination: global___LoggingConfig.BigQueryDestination | None = ...,
        redshift_destination: global___LoggingConfig.RedshiftDestination | None = ...,
        snowflake_destination: global___LoggingConfig.SnowflakeDestination | None = ...,
        custom_destination: global___LoggingConfig.CustomDestination | None = ...,
        athena_destination: global___LoggingConfig.AthenaDestination | None = ...,
        couchbase_columnar_destination: global___LoggingConfig.CouchbaseColumnarDestination | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["athena_destination", b"athena_destination", "bigquery_destination", b"bigquery_destination", "couchbase_columnar_destination", b"couchbase_columnar_destination", "custom_destination", b"custom_destination", "destination", b"destination", "file_destination", b"file_destination", "redshift_destination", b"redshift_destination", "snowflake_destination", b"snowflake_destination"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["athena_destination", b"athena_destination", "bigquery_destination", b"bigquery_destination", "couchbase_columnar_destination", b"couchbase_columnar_destination", "custom_destination", b"custom_destination", "destination", b"destination", "file_destination", b"file_destination", "redshift_destination", b"redshift_destination", "sample_rate", b"sample_rate", "snowflake_destination", b"snowflake_destination"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["destination", b"destination"]) -> typing_extensions.Literal["file_destination", "bigquery_destination", "redshift_destination", "snowflake_destination", "custom_destination", "athena_destination", "couchbase_columnar_destination"] | None: ...

global___LoggingConfig = LoggingConfig

class FeatureServiceList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURESERVICES_FIELD_NUMBER: builtins.int
    @property
    def featureservices(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FeatureService]: ...
    def __init__(
        self,
        *,
        featureservices: collections.abc.Iterable[global___FeatureService] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["featureservices", b"featureservices"]) -> None: ...

global___FeatureServiceList = FeatureServiceList



from dataclasses import dataclass
from typing import List, Optional, cast

import pandas as pd
import pyarrow

from feast.infra.common.materialization_job import (
    MaterializationJob,
    MaterializationJobStatus,
)
from feast.infra.compute_engines.dag.context import ExecutionContext
from feast.infra.compute_engines.dag.plan import ExecutionPlan
from feast.infra.compute_engines.local.arrow_table_value import ArrowTableValue
from feast.infra.offline_stores.offline_store import RetrievalJob, RetrievalMetadata
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.saved_dataset import SavedDatasetStorage


class LocalRetrievalJob(RetrievalJob):
    def __init__(
        self,
        plan: Optional[ExecutionPlan],
        context: ExecutionContext,
        full_feature_names: bool = True,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        metadata: Optional[RetrievalMetadata] = None,
        error: Optional[BaseException] = None,
    ):
        self._plan = plan
        self._context = context
        self._arrow_table = None
        self._error = error
        self._metadata = metadata
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = on_demand_feature_views or []

    def error(self) -> Optional[BaseException]:
        return self._error

    def _ensure_executed(self):
        if self._arrow_table is None:
            result = cast(ArrowTableValue, self._plan.execute(self._context))
            self._arrow_table = result.data

    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        self._ensure_executed()
        assert self._arrow_table is not None
        return self._arrow_table.to_pandas()

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pyarrow.Table:
        self._ensure_executed()
        return self._arrow_table

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: bool = False,
        timeout: Optional[int] = None,
    ):
        pass

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata

    def to_remote_storage(self) -> List[str]:
        raise NotImplementedError(
            "Remote storage is not supported in LocalRetrievalJob"
        )

    def to_sql(self) -> str:
        raise NotImplementedError(
            "SQL generation is not supported in LocalRetrievalJob"
        )


@dataclass
class LocalMaterializationJob(MaterializationJob):
    def __init__(
        self,
        job_id: str,
        status: MaterializationJobStatus,
        error: Optional[BaseException] = None,
    ) -> None:
        super().__init__()
        self._job_id: str = job_id
        self._status: MaterializationJobStatus = status
        self._error: Optional[BaseException] = error

    def status(self) -> MaterializationJobStatus:
        return self._status

    def error(self) -> Optional[BaseException]:
        return self._error

    def should_be_retried(self) -> bool:
        return False

    def job_id(self) -> str:
        return self._job_id

    def url(self) -> Optional[str]:
        return None



from feast.infra.offline_stores.contrib.couchbase_offline_store.tests.data_source import (
    CouchbaseColumnarDataSourceCreator,
)
from tests.integration.feature_repos.integration_test_repo_config import (
    IntegrationTestRepoConfig,
)
from tests.integration.feature_repos.repo_configuration import REDIS_CONFIG
from tests.integration.feature_repos.universal.online_store.redis import (
    RedisOnlineStoreCreator,
)

FULL_REPO_CONFIGS = [
    IntegrationTestRepoConfig(
        provider="aws",
        offline_store_creator=CouchbaseColumnarDataSourceCreator,
    ),
]

AVAILABLE_OFFLINE_STORES = [("aws", CouchbaseColumnarDataSourceCreator)]
AVAILABLE_ONLINE_STORES = {"redis": (REDIS_CONFIG, RedisOnlineStoreCreator)}



import logging
import multiprocessing

import click

from feast.constants import (
    DEFAULT_FEATURE_TRANSFORMATION_SERVER_PORT,
    DEFAULT_OFFLINE_SERVER_PORT,
    DEFAULT_REGISTRY_REST_SERVER_PORT,
    DEFAULT_REGISTRY_SERVER_PORT,
)
from feast.repo_operations import create_feature_store

logging.basicConfig(level=logging.INFO)


@click.command("serve")
@click.option(
    "--host",
    "-h",
    type=click.STRING,
    default="127.0.0.1",
    show_default=True,
    help="Specify a host for the server",
)
@click.option(
    "--port",
    "-p",
    type=click.INT,
    default=6566,
    show_default=True,
    help="Specify a port for the server",
)
@click.option(
    "--type",
    "-t",
    "type_",
    type=click.STRING,
    default="http",
    show_default=True,
    help="Specify a server type: 'http' or 'grpc'",
)
@click.option(
    "--no-access-log",
    is_flag=True,
    show_default=True,
    help="Disable the Uvicorn access log",
)
@click.option(
    "--workers",
    "-w",
    type=click.INT,
    default=1,
    show_default=True,
    help="Number of worker",
)
@click.option(
    "--keep-alive-timeout",
    type=click.INT,
    default=5,
    show_default=True,
    help="Timeout for keep alive",
)
@click.option(
    "--registry_ttl_sec",
    "-r",
    help="Number of seconds after which the registry is refreshed",
    type=click.INT,
    default=5,
    show_default=True,
)
@click.option(
    "--key",
    "-k",
    "tls_key_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS certificate private key. You need to pass --cert as well to start server in TLS mode",
)
@click.option(
    "--cert",
    "-c",
    "tls_cert_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS certificate public key. You need to pass --key as well to start server in TLS mode",
)
@click.option(
    "--metrics",
    "-m",
    is_flag=True,
    show_default=True,
    help="Enable the Metrics Server",
)
@click.pass_context
def serve_command(
    ctx: click.Context,
    host: str,
    port: int,
    type_: str,
    no_access_log: bool,
    workers: int,
    metrics: bool,
    keep_alive_timeout: int,
    tls_key_path: str,
    tls_cert_path: str,
    registry_ttl_sec: int = 5,
):
    """Start a feature server locally on a given port."""
    if (tls_key_path and not tls_cert_path) or (not tls_key_path and tls_cert_path):
        raise click.BadParameter(
            "Please pass --cert and --key args to start the feature server in TLS mode."
        )
    store = create_feature_store(ctx)

    store.serve(
        host=host,
        port=port,
        type_=type_,
        no_access_log=no_access_log,
        workers=workers,
        metrics=metrics,
        keep_alive_timeout=keep_alive_timeout,
        tls_key_path=tls_key_path,
        tls_cert_path=tls_cert_path,
        registry_ttl_sec=registry_ttl_sec,
    )


@click.command("serve_transformations")
@click.option(
    "--port",
    "-p",
    type=click.INT,
    default=DEFAULT_FEATURE_TRANSFORMATION_SERVER_PORT,
    help="Specify a port for the server",
)
@click.pass_context
def serve_transformations_command(ctx: click.Context, port: int):
    """[Experimental] Start a feature consumption server locally on a given port."""
    store = create_feature_store(ctx)

    store.serve_transformations(port)


@click.command("serve_registry")
@click.option(
    "--port",
    "-p",
    type=click.INT,
    default=DEFAULT_REGISTRY_SERVER_PORT,
    help="Specify a port for the server",
)
@click.option(
    "--rest-port",
    type=click.INT,
    default=DEFAULT_REGISTRY_REST_SERVER_PORT,
    show_default=True,
    help="Specify a port for the REST API server (if enabled).",
)
@click.option(
    "--key",
    "-k",
    "tls_key_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS certificate private key. You need to pass --cert as well to start server in TLS mode",
)
@click.option(
    "--cert",
    "-c",
    "tls_cert_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS certificate public key. You need to pass --key as well to start server in TLS mode",
)
@click.option(
    "--grpc/--no-grpc",
    is_flag=True,
    default=True,
    show_default=True,
    help="Start a gRPC Registry Server. Enabled by default.",
)
@click.option(
    "--rest-api",
    "-r",
    is_flag=True,
    default=False,
    show_default=True,
    help="Start a REST API Registry Server",
)
@click.pass_context
def serve_registry_command(
    ctx: click.Context,
    port: int,
    tls_key_path: str,
    tls_cert_path: str,
    grpc: bool,
    rest_api: bool,
    rest_port: int,
):
    """Start Feast Registry server (gRPC by default, REST opt-in)."""
    if (tls_key_path and not tls_cert_path) or (not tls_key_path and tls_cert_path):
        raise click.BadParameter(
            "Please pass --cert and --key args to start the registry server in TLS mode."
        )
    store = create_feature_store(ctx)

    if grpc and rest_api:
        repo_path = store.repo_path
        multiprocessing.set_start_method("spawn", force=True)
        servers = [
            multiprocessing.Process(
                target=_serve_grpc_registry,
                args=(repo_path, port, tls_key_path, tls_cert_path),
                name="grpc_registry_server",
            ),
            multiprocessing.Process(
                target=_serve_rest_registry,
                args=(repo_path, rest_port, tls_key_path, tls_cert_path),
                name="rest_registry_server",
            ),
        ]
        logging.info("Starting Feast Registry servers (gRPC + REST)...")
        for p in servers:
            logging.info(f"Starting {p.name}")
            p.start()
        for p in servers:
            p.join()
    else:
        if grpc:
            store.serve_registry(port, tls_key_path, tls_cert_path)
        else:
            store.serve_registry(
                port=rest_port,
                tls_key_path=tls_key_path,
                tls_cert_path=tls_cert_path,
                rest_api=rest_api,
            )


def _serve_grpc_registry(
    repo_path: str, port: int, tls_key_path: str, tls_cert_path: str
):
    from feast import FeatureStore

    store = FeatureStore(repo_path=repo_path)
    store.serve_registry(
        port=port,
        tls_key_path=tls_key_path,
        tls_cert_path=tls_cert_path,
    )


def _serve_rest_registry(
    repo_path: str, port: int, tls_key_path: str, tls_cert_path: str
):
    from feast import FeatureStore

    store = FeatureStore(repo_path=repo_path)
    store.serve_registry(
        port=port,
        tls_key_path=tls_key_path,
        tls_cert_path=tls_cert_path,
        rest_api=True,
    )


@click.command("serve_offline")
@click.option(
    "--host",
    "-h",
    type=click.STRING,
    default="127.0.0.1",
    show_default=True,
    help="Specify a host for the server",
)
@click.option(
    "--port",
    "-p",
    type=click.INT,
    default=DEFAULT_OFFLINE_SERVER_PORT,
    help="Specify a port for the server",
)
@click.option(
    "--key",
    "-k",
    "tls_key_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS certificate private key. You need to pass --cert as well to start server in TLS mode",
)
@click.option(
    "--cert",
    "-c",
    "tls_cert_path",
    type=click.STRING,
    default="",
    show_default=False,
    help="path to TLS certificate public key. You need to pass --key as well to start server in TLS mode",
)
@click.pass_context
def serve_offline_command(
    ctx: click.Context,
    host: str,
    port: int,
    tls_key_path: str,
    tls_cert_path: str,
):
    """Start a remote server locally on a given host, port."""
    if (tls_key_path and not tls_cert_path) or (not tls_key_path and tls_cert_path):
        raise click.BadParameter(
            "Please pass --cert and --key args to start the offline server in TLS mode."
        )
    store = create_feature_store(ctx)

    store.serve_offline(host, port, tls_key_path, tls_cert_path)



# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Dict, List, TypeAlias, Union

import pandas as pd
import pyarrow as pa

from feast.feature_view import DUMMY_ENTITY_ID
from feast.protos.feast.serving.ServingService_pb2 import GetOnlineFeaturesResponse
from feast.torch_wrapper import get_torch
from feast.type_map import feast_value_type_to_python_type

if TYPE_CHECKING:
    import torch

    TorchTensor: TypeAlias = torch.Tensor
else:
    TorchTensor: TypeAlias = Any

TIMESTAMP_POSTFIX: str = "__ts"


class OnlineResponse:
    """
    Defines an online response in feast.
    """

    def __init__(self, online_response_proto: GetOnlineFeaturesResponse):
        """
        Construct a native online response from its protobuf version.

        Args:
        online_response_proto: GetOnlineResponse proto object to construct from.
        """
        self.proto = online_response_proto
        # Delete DUMMY_ENTITY_ID from proto if it exists
        for idx, val in enumerate(self.proto.metadata.feature_names.val):
            if val == DUMMY_ENTITY_ID:
                del self.proto.metadata.feature_names.val[idx]
                del self.proto.results[idx]

                break

    def to_dict(self, include_event_timestamps: bool = False) -> Dict[str, Any]:
        """
        Converts GetOnlineFeaturesResponse features into a dictionary form.

        Args:
        include_event_timestamps: bool Optionally include feature timestamps in the dictionary
        """
        response: Dict[str, List[Any]] = {}

        for feature_ref, feature_vector in zip(
            self.proto.metadata.feature_names.val, self.proto.results
        ):
            response[feature_ref] = [
                feast_value_type_to_python_type(v) for v in feature_vector.values
            ]

            if include_event_timestamps:
                timestamp_ref = feature_ref + TIMESTAMP_POSTFIX
                response[timestamp_ref] = [
                    ts.seconds for ts in feature_vector.event_timestamps
                ]

        return response

    def to_df(self, include_event_timestamps: bool = False) -> pd.DataFrame:
        """
        Converts GetOnlineFeaturesResponse features into Panda dataframe form.

        Args:
        include_event_timestamps: bool Optionally include feature timestamps in the dataframe
        """

        return pd.DataFrame(self.to_dict(include_event_timestamps))

    def to_arrow(self, include_event_timestamps: bool = False) -> pa.Table:
        """
        Converts GetOnlineFeaturesResponse features into pyarrow Table.

        Args:
        include_event_timestamps: bool Optionally include feature timestamps in the table
        """

        return pa.Table.from_pydict(self.to_dict(include_event_timestamps))

    def to_tensor(
        self,
        kind: str = "torch",
        default_value: Any = float("nan"),
    ) -> Dict[str, Union[TorchTensor, List[Any]]]:
        """
        Converts GetOnlineFeaturesResponse features into a dictionary of tensors or lists.

        - Numeric features (int, float, bool) -> torch.Tensor
        - Non-numeric features (e.g., strings) -> list[Any]

        Args:
            kind: Backend tensor type. Currently only "torch" is supported.
            default_value: Value to substitute for missing (None) entries.

        Returns:
            Dict[str, Union[torch.Tensor, List[Any]]]: Mapping of feature names to tensors or lists.
        """
        if kind != "torch":
            raise ValueError(
                f"Unsupported tensor kind: {kind}. Only 'torch' is supported currently."
            )
        torch = get_torch()
        feature_dict = self.to_dict(include_event_timestamps=False)
        feature_keys = set(self.proto.metadata.feature_names.val)
        tensor_dict: Dict[str, Union[TorchTensor, List[Any]]] = {}
        for key in feature_keys:
            raw_values = feature_dict[key]
            values = [v if v is not None else default_value for v in raw_values]
            first_valid = next((v for v in values if v is not None), None)
            if isinstance(first_valid, (int, float, bool)):
                try:
                    device = "cuda" if torch.cuda.is_available() else "cpu"
                    tensor_dict[key] = torch.tensor(values, device=device)
                except Exception as e:
                    raise ValueError(
                        f"Failed to convert values for '{key}' to tensor: {e}"
                    )
            else:
                tensor_dict[key] = (
                    values  # Return as-is for strings or unsupported types
                )
        return tensor_dict



"""Configuration for Ray compute engine."""

from datetime import timedelta
from typing import Any, Dict, Literal, Optional

from pydantic import StrictStr

from feast.repo_config import FeastConfigBaseModel


class RayComputeEngineConfig(FeastConfigBaseModel):
    """Configuration for Ray Compute Engine."""

    type: Literal["ray.engine"] = "ray.engine"
    """Ray Compute Engine type selector"""

    ray_address: Optional[str] = None
    """Ray cluster address. If None, uses local Ray cluster."""

    staging_location: Optional[StrictStr] = None
    """Remote path for batch materialization jobs"""

    # Ray-specific performance configurations
    broadcast_join_threshold_mb: int = 100
    """Threshold for using broadcast joins (in MB)"""

    enable_distributed_joins: bool = True
    """Whether to enable distributed joins for large datasets"""

    max_parallelism_multiplier: int = 2
    """Multiplier for max parallelism based on available CPUs"""

    target_partition_size_mb: int = 64
    """Target partition size in MB"""

    window_size_for_joins: str = "1H"
    """Window size for windowed temporal joins"""

    ray_conf: Optional[Dict[str, Any]] = None
    """Ray configuration parameters"""

    # Additional configuration options
    max_workers: Optional[int] = None
    """Maximum number of Ray workers. If None, uses all available cores."""

    enable_optimization: bool = True
    """Enable automatic performance optimizations."""

    execution_timeout_seconds: Optional[int] = None
    """Timeout for job execution in seconds."""

    @property
    def window_size_timedelta(self) -> timedelta:
        """Convert window size string to timedelta."""
        if self.window_size_for_joins.endswith("H"):
            hours = int(self.window_size_for_joins[:-1])
            return timedelta(hours=hours)
        elif self.window_size_for_joins.endswith("min"):
            minutes = int(self.window_size_for_joins[:-3])
            return timedelta(minutes=minutes)
        elif self.window_size_for_joins.endswith("s"):
            seconds = int(self.window_size_for_joins[:-1])
            return timedelta(seconds=seconds)
        else:
            # Default to 1 hour
            return timedelta(hours=1)



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import numpy as np

from feast import FeatureStore, FeatureView
from feast.online_response import OnlineResponse


class FeastVectorStore:
    """
    Feast-based vector store implementation with support for text, image, and multi-modal search.
    This class provides a convenient interface for vector similarity search using Feast's
    vector database integrations. It supports:
    - Text similarity search using text embeddings
    - Image similarity search using image embeddings
    - Multi-modal search combining text and image queries
    """

    def __init__(self, repo_path: str, rag_view: FeatureView, features: list[str]):
        """Initialize the Feast vector store.

        Args:
            repo_path: Path to the Feast repo
            rag_view: Feature view
            features: List of feature names to retrieve
        """
        self._store = None  # Lazy load
        self._store_repo_path = repo_path
        self.rag_view = rag_view
        self.features = features

    @property
    def store(self):
        if self._store is None:
            self._store = FeatureStore(repo_path=self._store_repo_path)
            self._store.apply([self.rag_view])
            # TODO: Add validation to ensure store type is one of supported types e.g. pgvector, elasticsearch, milvus
        return self._store

    def query(
        self,
        query_vector: Optional[np.ndarray] = None,
        query_string: Optional[str] = None,
        query_image_bytes: Optional[bytes] = None,
        top_k: int = 10,
    ) -> OnlineResponse:
        """Query the Feast vector store with support for text, image, and multi-modal search.

        Args:
            query_vector: Optional vector to use for similarity search (text embeddings)
            query_string: Optional string query for keyword/semantic search
            query_image_bytes: Optional image bytes for image similarity search
            top_k: Number of results to return

        Returns:
            An OnlineResponse
        """
        query_list = query_vector.tolist() if query_vector is not None else None

        distance_metric = None
        for field in self.rag_view.schema:
            if hasattr(field, "vector_index") and field.vector_index:
                if hasattr(field, "vector_search_metric"):
                    distance_metric = field.vector_search_metric
                    break

        return self.store.retrieve_online_documents_v2(
            features=self.features,
            query=query_list,
            query_string=query_string,
            query_image_bytes=query_image_bytes,
            top_k=top_k,
            distance_metric=distance_metric,
        )



import logging
import os

import jwt
import requests

from feast.permissions.auth_model import OidcClientAuthConfig
from feast.permissions.client.auth_client_manager import AuthenticationClientManager
from feast.permissions.oidc_service import OIDCDiscoveryService

logger = logging.getLogger(__name__)


class OidcAuthClientManager(AuthenticationClientManager):
    def __init__(self, auth_config: OidcClientAuthConfig):
        self.auth_config = auth_config

    def get_token(self):
        intra_communication_base64 = os.getenv("INTRA_COMMUNICATION_BASE64")
        # If intra server communication call
        if intra_communication_base64:
            payload = {
                "preferred_username": f"{intra_communication_base64}",  # Subject claim
            }

            return jwt.encode(payload, "")

        # Fetch the token endpoint from the discovery URL
        token_endpoint = OIDCDiscoveryService(
            self.auth_config.auth_discovery_url
        ).get_token_url()

        # 1) pre-issued JWT supplied in config
        if getattr(self.auth_config, "token", None):
            return self.auth_config.token

        # 2) client_credentials
        if self.auth_config.client_secret and not (
            self.auth_config.username and self.auth_config.password
        ):
            token_request_body = {
                "grant_type": "client_credentials",
                "client_id": self.auth_config.client_id,
                "client_secret": self.auth_config.client_secret,
            }
        # 3) ROPG (username + password + client_secret)
        else:
            token_request_body = {
                "grant_type": "password",
                "client_id": self.auth_config.client_id,
                "client_secret": self.auth_config.client_secret,
                "username": self.auth_config.username,
                "password": self.auth_config.password,
            }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        token_response = requests.post(
            token_endpoint, data=token_request_body, headers=headers
        )
        if token_response.status_code == 200:
            access_token = token_response.json()["access_token"]
            if not access_token:
                logger.debug(
                    f"access_token is empty for the client_id=${self.auth_config.client_id}"
                )
                raise RuntimeError("access token is empty")
            return access_token
        else:
            raise RuntimeError(
                f"""Failed to obtain oidc access token:url=[{token_endpoint}] {token_response.status_code} - {token_response.text}"""
            )



import click
import yaml

from feast import utils
from feast.cli.cli_options import tagsOption
from feast.errors import FeastObjectNotFoundException
from feast.repo_operations import create_feature_store


@click.group(name="projects")
def projects_cmd():
    """
    Access projects
    """
    pass


@projects_cmd.command("describe")
@click.argument("name", type=click.STRING)
@click.pass_context
def project_describe(ctx: click.Context, name: str):
    """
    Describe a project
    """
    store = create_feature_store(ctx)

    try:
        project = store.get_project(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(project)), default_flow_style=False, sort_keys=False
        )
    )


@projects_cmd.command("current_project")
@click.pass_context
def project_current(ctx: click.Context):
    """
    Returns the current project configured with FeatureStore object
    """
    store = create_feature_store(ctx)

    try:
        project = store.get_project(name=None)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(project)), default_flow_style=False, sort_keys=False
        )
    )


@projects_cmd.command(name="list")
@tagsOption
@click.pass_context
def project_list(ctx: click.Context, tags: list[str]):
    """
    List all projects
    """
    store = create_feature_store(ctx)
    table = []
    tags_filter = utils.tags_list_to_dict(tags)
    for project in store.list_projects(tags=tags_filter):
        table.append([project.name, project.description, project.tags, project.owner])

    from tabulate import tabulate

    print(
        tabulate(
            table, headers=["NAME", "DESCRIPTION", "TAGS", "OWNER"], tablefmt="plain"
        )
    )
